/*************************************************************************\
**  source       * dbe0lb.c
**  directory    * dbe
**  description  * Log buffers
**               * 
**               * Copyright (C) 2006 Solid Information Technology Ltd
\*************************************************************************/
/*
  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; only under version 2 of the License.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
  USA
*/

#ifdef DOCUMENTATION
**************************************************************************

Implementation:
--------------

Limitations:
-----------

Error handling:
--------------


Objects used:
------------


Preconditions:
-------------


Multithread considerations:
--------------------------

Example:
-------


**************************************************************************
#endif /* DOCUMENTATION */

#include <ssc.h>
#include <sssem.h>
#include <ssmem.h>
#include <ssdebug.h>
#include <ssstring.h>
#include <sssprint.h>
#include <sslimits.h>
#include <ssfile.h>
#include <ssltoa.h>
#include <sspmon.h>

#include "dbe0lb.h"

#ifndef SS_NOLOGGING

void dbe_lb_initbuf(dbe_logbuf_t* logbuf, size_t bufsize)
{
        dbe_logbuf_t* lb;
        logbuf->lb_.ctr[LB_VNUM_IDX] = (ss_byte_t)1;
        logbuf->lb_.ctr[LB_BNUM_IDX] = (ss_byte_t)1;
        lb = LB_CHK2ADDR(logbuf, bufsize);
        LB_CP_CHK(lb, logbuf);
}
        
/*#***********************************************************************\
 * 
 *		dbe_lb_init
 * 
 * Creates logbuf object
 * 
 * Parameters : 
 * 
 *	bufsize - in
 *		buffer size in bytes (multiple of cluster size)
 *		
 * Return value - give:
 *      pointer to created object
 * 
 * Limitations  : 
 * 
 * Globals used : 
 */
dbe_logbuf_t* dbe_lb_init(size_t bufsize)
{
        dbe_logbuf_t* logbuf;

        logbuf = SsMemAlloc(bufsize);
        dbe_lb_initbuf(logbuf, bufsize);
        ss_pprintf_1(("dbe_lb_init:logbuf %x\n", logbuf));
        return (logbuf);
}

/*#***********************************************************************\
 * 
 *		dbe_lb_done
 * 
 * Deletes a logbuf object
 * 
 * Parameters : 
 * 
 *	logbuf - in, take
 *		pointer to logbuf
 *		
 * Return value : 
 * 
 * Limitations  : 
 * 
 * Globals used : 
 */
void dbe_lb_done(dbe_logbuf_t* logbuf)
{
        ss_pprintf_1(("dbe_lb_done:logbuf %x\n", logbuf));
        SsMemFree(logbuf);
}

/*#***********************************************************************\
 *
 *      dbe_lb_isconsistent
 *
 * Checks whether a log file block is completely written to disk.
 *
 * Parameters :
 *
 *  logbuf - in, use
 *      pointer to logbuf
 *
 *  bufsize - in
 *      buf size in bytes
 *
 * Return value :
 *      TRUE when ok or
 *      FALSE when block is corrupt
 *
 * Limitations  :
 *
 * Globals used :
 */
bool dbe_lb_isconsistent(dbe_logbuf_t* logbuf, size_t bufsize)
{
        dbe_logbuf_t* lb;
        lb = LB_CHK2ADDR(logbuf, bufsize);

        return (lb->lb_.ctr[LB_BNUM_IDX] == logbuf->lb_.ctr[LB_BNUM_IDX]
             && lb->lb_.ctr[LB_VNUM_IDX] == logbuf->lb_.ctr[LB_VNUM_IDX]);
}

#endif /* SS_NOLOGGING */
