/*************************************************************************\
**  source       * ssenv.h
**  directory    * ss
**  description  * OS and environment specific compiler options
**               * and definitions.
**               * This file should be included before all other 
**               * headers.
**               * 
**               * HOW TO PORT SOLID TO A NEW ENVIRONMENT
**               * 
**               * At least the following source files need to be modified
**               * when porting solid to a new environment:
**               * 
**               *      ssenv.?
**               *      sslimits.h
**               *      ssfloat.h
**               *      sstraph.c
**               *      sstraph2.c
**               *      sssem???.c
**               *      ssc.h       
**               *      ssfile.?    (non-unix)  
**               *      ssproces.?  (non-unix)  
**               *      ses*.c     (possible communication changes)
**               *      com0pdef.c (default listen/connect infos)
**               *      com0cfg.h  (non-unix: default msgsizes)
**               *      css/rguti.c (filename conversion env)
**               *      sse/solinst.c (unixes)
**               * 
**               * The following make-environment files need also
**               * modifications:
**               * 
**               *      root.inc
**               *          new makefile define for the OS
**               *          add also to so/root.inc
**               *      makefile.inc
**               *          branch for the OS
**               *          new command line define for OS
**               * 
**               * Copyright (C) 2006 Solid Information Technology Ltd
\*************************************************************************/
/*
  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; only under version 2 of the License.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
  USA
*/

#ifndef SSENV_H
#define SSENV_H

#include "ssmyenv.h"

#if 0 /* Removed 2006-10-18 apl */
#define SS_FORCE_NONVERSCOUNTED2SSMEMALLOC /* !!! temporary !!! */
#endif

#ifndef SS_TC_CLIENT
#define SS_TC_CLIENT
#endif

#ifndef SS_MME
#define SS_MME
/* The following are mandatory 2006-10-23 apl */
#define SS_MMEG2
#define MMEG2_MUTEXING
#define MMEG2_VERSIONING
#define MMEG2_SHARED_READS
#define SS_BNODE_INTEGRATED
#define SS_BNODE_JOIN
#define SS_FFMEM_FREELEVELS
#define LMGR_SPLITMUTEX
#endif

#ifndef SS_NOMMINDEX
#define SS_NOMMINDEX
#endif

#ifndef SS_TASKOPTIMIZE
#define SS_TASKOPTIMIZE
#endif

#ifndef SS_ACCOPTIMIZE
#define SS_ACCOPTIMIZE
#endif

#ifndef SS_NETOPTIMIZE
#define SS_NETOPTIMIZE
#endif

#ifdef ENABLE_HISTORY_OPTIMIZATION
#undef ENABLE_HISTORY_OPTIMIZATION
#endif

#ifndef SS_HSBG2
#define SS_HSBG2
#endif

#ifndef RS_FLATAVAL
#define RS_FLATAVAL
#endif

#if !defined(SS_FFMEM) && !defined(SS_MYSQL)
#define SS_FFMEM
#endif

#if !defined(SS_MYSQL)
#if !defined(__attribute__) && (defined(__cplusplus) || !defined(__GNUC__)  || __GNUC__ == 2 && __GNUC_MINOR__ < 8)
#define __attribute__(A)
#endif
#endif /* !SS_MYSQL */

#ifndef SS_COREOPT
#define SS_COREOPT
#endif

#ifdef SS_DEBUG
# ifndef SS_BETA
#  define SS_BETA
# endif /* SS_BETA */
# ifndef SS_PROFILE
#  define SS_PROFILE
# endif /* SS_PROFILE */
#endif /* SS_DEBUG */

#if defined(SS_SEMPROFILE) && !defined(SS_PROFILE)
# define SS_PROFILE
#endif
#if defined(SS_PROFILE) && !defined(SS_SEMPROFILE)
# define SS_SEMPROFILE
#endif

#if defined(SS_DEBUG) || defined(SS_BETA) || defined(SS_PURIFY) || defined(SS_COVER)
# undef SS_HSBG2_TESTGA
#else 
# define SS_HSBG2_TESTGA
#endif

#define HSB_LPID
#define HSBG2_NEW_BUSY_DETECT
#define HSBG2_ASYNC_REMOTE_DURABLE_ACK

#ifndef EASE_OF_USE_EVENTS
#define EASE_OF_USE_EVENTS
#endif

#ifndef COLLATION_UPDATE
#define COLLATION_UPDATE
#endif

#ifndef SS_ALTER_TABLE
#define SS_ALTER_TABLE
#endif

#ifndef REFERENTIAL_INTEGRITY
#define REFERENTIAL_INTEGRITY
#endif

#ifndef RS_REFTVALS
#define RS_REFTVALS
#endif

#ifndef SS_REFDVA_MEMLINK
#define SS_REFDVA_MEMLINK
#endif

#ifndef LICENSEINFO_V2
#define LICENSEINFO_V2
#endif

#ifndef SS_SYNC
#define SS_SYNC
#endif

#ifdef SS_UNICODE_ALL
#define SS_UNICODE_DATA
#define SS_UNICODE_SQL
#define SS_UNICODE_CLI
#endif

#ifndef SS_CATALOGSUPP
#define SS_CATALOGSUPP
#endif

#ifndef DBE_REPLICATION
# define DBE_REPLICATION
#endif

#ifndef DBE_HSB_REPLICATION
# define DBE_HSB_REPLICATION
#endif

#ifndef HSB_FAILURE_FIX
# define HSB_FAILURE_FIX
#endif

#ifndef DBE_LOGORDERING_FIX
# define DBE_LOGORDERING_FIX
#endif

#ifndef DBE_BNODE_MISMATCHARRAY
# define DBE_BNODE_MISMATCHARRAY
#endif

/* SOLIDFAST means direct function calls in SQL without function pointer table. */
#ifndef SOLIDFAST
# define SOLIDFAST
#endif

#ifndef SS_FILEPATHPREFIX
# define SS_FILEPATHPREFIX
#endif

#ifndef SS_BRKSOCKET_WITH_SOCKETPAIR
# define SS_BRKSOCKET_WITH_SOCKETPAIR
#endif

#ifndef SS_CONBLOCK
# define SS_CONBLOCK
#endif

typedef enum SsCpuIdEnum {

        SS_CPU_NULL         = 0,
        SS_CPU_IX86         = 1,
        SS_CPU_VAX          = 2,
        SS_CPU_ALPHA        = 3,
        SS_CPU_POWERPC      = 4,
        SS_CPU_PA_RISC      = 5,
        SS_CPU_SPARC        = 6,
        SS_CPU_MIPS         = 7,
        SS_CPU_CLIPPER      = 8,
        SS_CPU_ARM          = 9,
        SS_CPU_VXSIMHP      = 10,
        SS_CPU_VXSIMSOLARIS = 11,
        SS_CPU_MIPSII       = 12,
        SS_CPU_IA64	        = 13,
	SS_CPU_S390         = 14,
	SS_CPU_XSCALE_BE    = 15,
        SS_CPU_AMD64	    = 16,
	SS_CPU_SH4_BE       = 17,
	SS_CPU_SH4_LE       = 18,
	SS_CPU_SPARC64      = 19,
        SS_CPU_PA_RISC64    = 20,
        SS_CPU_IA64_32BIT   = 21,
        SS_CPU_ENDVALUE

} SsCpuIdT;
/*
 * Note: SsCpuIdEnum, SsOsIdEnum and SsEnvIdEnum must have exactly the
 * same value across all branches and versions. 
 */

typedef enum SsOsIdEnum {

        SS_OS_NULL      = 0,

        SS_OS_DOS       = 1,
        SS_OS_WIN       = 2,
        SS_OS_W95       = 3, /* W95 has osvers 4.0, W98 has osvers 4.10 */
        SS_OS_WNT       = 4,
        SS_OS_OS2       = 5,
        SS_OS_VMS       = 6,
        SS_OS_DECUNIX   = 7,
        SS_OS_AIX       = 8,
        SS_OS_SCOUNIX   = 9,
        SS_OS_NW        = 10,
        SS_OS_LINUX     = 11,
        SS_OS_SOLARIS   = 12,
        SS_OS_HPUX      = 13,
        SS_OS_SVR4      = 14,
        SS_OS_CLIX      = 15,
        SS_OS_IRIX      = 16,
        SS_OS_WCE       = 17,
        SS_OS_SUW       = 18,
        SS_OS_VXW       = 19,
        SS_OS_FBX       = 20,
        SS_OS_CRX       = 21,
        SS_OS_E32       = 22,
        SS_OS_PSOS      = 23,
        SS_OS_OSE       = 24,
	SS_OS_QNX       = 25,
        SS_OS_BSDI      = 26,

        SS_OS_ENDVALUE

} SsOsIdT;
/*
 * Note: SsCpuIdEnum, SsOsIdEnum and SsEnvIdEnum must have exactly the
 * same value across all branches and versions. 
 */

typedef enum SsEnvIdEnum {

        SS_ENVID_DOS = 0,
        SS_ENVID_D4G,
        SS_ENVID_W16 = 10,
        SS_ENVID_W95 = 11,
        SS_ENVID_NTI = 12,
        SS_ENVID_NTA = 13,
        SS_ENVID_WCE = 14,
        SS_ENVID_W98 = 15,
        SS_ENVID_O16 = 20,
        SS_ENVID_O32,
        SS_ENVID_NW3 = 30,
        SS_ENVID_NW4,
        SS_ENVID_OVV = 40,
        SS_ENVID_OVA,
        SS_ENVID_A3X = 50,
        SS_ENVID_A4X,
        SS_ENVID_A4X64BIT,
        SS_ENVID_A5X,
        SS_ENVID_H9X = 60,
        SS_ENVID_H0X,
        SS_ENVID_H1X,
        SS_ENVID_H1X64BIT,
        SS_ENVID_HIA,
        SS_ENVID_HIA64BIT,
        SS_ENVID_SCX = 70,
        SS_ENVID_SSX = 80,
        SS_ENVID_S8X = 81,
        SS_ENVID_S8X64BIT   = 82,
        SS_ENVID_S9X        = 83,
        SS_ENVID_S9X64BIT   = 84,
        SS_ENVID_S0X        = 85,
        SS_ENVID_S0X64BIT   = 86,
        SS_ENVID_S0XI       = 87,
        SS_ENVID_S0XI64BIT  = 88,
        SS_ENVID_IRX = 90,
        SS_ENVID_L2X64 = 95,
        SS_ENVID_LUX = 100,
        SS_ENVID_LXA = 101,
        SS_ENVID_L2X = 102,
        SS_ENVID_LPX = 103,
        SS_ENVID_LSAX = 104,
	SS_ENVID_LS390X = 105,
	SS_ENVID_LXSB = 106,
	SS_ENVID_LSH4BX = 107,
	SS_ENVID_LSH4LX = 108,
        SS_ENVID_CLX = 110,
        SS_ENVID_DIX = 120,
        SS_ENVID_SUW = 130,
        SS_ENVID_VXW = 140,
        SS_ENVID_VPX,
	SS_ENVID_VMX,
        SS_ENVID_VSSX = 149,
        SS_ENVID_FBX = 150,
        SS_ENVID_FEX = 151,
        SS_ENVID_FEX64 = 152,
        SS_ENVID_CRX = 160,
        SS_ENVID_E32 = 170,
        SS_ENVID_PSP = 180,
	SS_ENVID_OSA = 190,
        SS_ENVID_OSE = 191,
        SS_ENVID_BSI = 192,
	SS_ENVID_QPX = 199,
	SS_ENVID_QNX = 200,
        SS_ENVID_WCE4M2 = 201,
        SS_ENVID_WNT64 = 202

} SsEnvIdT;
/*
 * Note: SsCpuIdEnum, SsOsIdEnum and SsEnvIdEnum must have exactly the
 * same value across all branches and versions. 
 */

/* ---------------------------------------------------------------------
 *          Version number information
 */
#ifndef SS_SERVER_VERSION
#define SS_SERVER_VERSION   ss_versiontext()
#endif

#ifndef SS_SERVER_NAME
#define SS_SERVER_NAME      ss_servername()
#endif

#define SS_GENERIC_DBMS_NAME "Solid Database Engine"
/*NOTE: generic name is also hardcoded in ssver.c, ssversndl.c, ssversnsf.c 
sssncver.c, sssereesf.c and makefile.inc */
#define SS_SERVER_VERSNUM   ss_versionnumber()
#define SS_SOLIDVERS        200

#define SS_SERVER_VERSNUM_MINOR     ss_vers_minor()
#define SS_SERVER_VERSNUM_MAJOR     ss_vers_major()
#define SS_SERVER_VERSNUM_RELEASE   ss_vers_release()

/* Copyright string. When you change this, change also makefile.inc.
 */
#define SS_COPYRIGHT        ss_copyright


/* The following variables are in ssdebug.c (to avoid linking problems).
 */
extern const char* ss_company_name;
extern const char* ss_copyright_short;
extern const char* ss_copyright;
extern char* ss_cmdline;
extern int   ss_migratehsbg2; /* bool */
extern int   ss_convertdb;    /* bool */

char*   ss_servername(void);
void    ss_setservername(char* name);
char*   ss_versiontext(void);
int     ss_versionnumber(void);
int     ss_vers_major(void);
int     ss_vers_minor(void);
int     ss_vers_release(void);
long    ss_codebaseversion(void);
int     ss_vers_dll(void);
char*   ss_vers_dllpostfix(void);
int     ss_vers_issync(void);
int     ss_vers_isaccelerator(void);
int     ss_vers_isdiskless(void);

/* ---------------------------------------------------------------------
 *          Operating system dependent part                         
 *
 *  You have to specify exactly one of the following operating systems
 *
 *      - SS_OS2
 *      - SS_WIN
 *      - SS_DOS
 *      - SS_UNIX
 *      - SS_AIX
 *      - SS_SCO
 *      - SS_LINUX
 *      - SS_HPUX
 *      - SS_NLM
 *      - SS_SVR4
 *      - SS_CLIX
 *      - SS_IRIX
 *
 *      Example: cc -c -DSS_AIX -DSS_UNIX
 *
 *
 *  SS_OS2 and SS_WIN can also contain an SS_DLL flag. It indicates that 
 *
 *      - All functions in source files containing an SS_EXPORT_C
 *        attribute are going to be exported from the becoming
 *        DLL module.
 *
 *      - All external functions containing an SS_EXPORT_H attribute in
 *        their prototype are located in a DLL.
 *        This means that in WINDOWS they can not be called from the same
 *        module without SsMakeCallbackFun (MakeProcInstance)
 *
 *      Example: cl -c -DSS_OS2 -DSS_DLL
 *
 *
 *  For test executables in DLL environment you can define SS_EXE to get
 *  the correct view to the system headers. (scanf in WINDOWS etc.)
 *  The combination SS_DLL and SS_EXE still expands SS_EXPORT to _export.
 *  It also defines WINVER 0x0300 in sswindow.h. The executables can therefore
 *  be run in WIN OS/2 2.0 session.
 *
 */

#if defined(SS_OS2)

#  define SS_MT
#  define OS2 SS_OS2

#  if defined (SS_DLL) && !defined(SS_EXE)
#    define DLL
#    define _DLL
#  endif /* SS_DLL */

#  if OS2 == 32
#    define SS_PMEM
#    define SS_ZMEM
#  endif

#  if defined(SS_XREF)
#    define _Optlink
#    define _Builtin
#  endif /* SS_XREF */

#elif defined(SS_W16)

#  define SS_ZMEM
#  define WINDOWS
#  ifndef _WINDOWS
#    define _WINDOWS
#  endif/* _WINDOWS */
#  if defined (SS_DLL) && !defined(SS_EXE)
#    define _WINDLL
#  endif /* SS_DLL */

#elif defined(SS_NT)

#define SS_LARGEFILE
#ifdef SS_DEBUG
#  define SS_ZMEM
#else /* SS_DEBUG */
#  define SS_FLATSEM /* enable flat+inline mutexes */
#endif /* SS_DEBUG */

#  define SS_PAGED_MEMALLOC_AVAILABLE
#ifdef SS_NEW_ST
#  define SS_SMALLSYSTEM
#else
#  define SS_MT
#endif

#if !defined(SS_USE_WIN32TLS_API) && defined(SS_MT)
#define SS_USE_WIN32TLS_API
#endif

#  if defined(SS_ALPHA)
#    ifndef WIN32
#      define WIN32
#    endif
#  endif

/* NT has UNICODE main function support */
#define SS_WMAIN_CAPABLE

#elif defined(SS_NLM)

/* #  define SS_MT */ /* Ari removed Dec 02, 1996 */
#  define SS_ZMEM

#elif defined (SS_DOS)

#  define DOS

#elif defined (SS_DOSX)

#  define SS_DOS
#  define DOS

#elif defined (SS_UNIX)

#  define UNIX

#endif

/* ---------------------------------------------------------------------
 *          Compiler dependent part                         
 *
 * You can define SS_NO_ANSI.
 *
 * If SS_NO_ANSI is NOT defined, the default is always SS_ANSI.
 * SS_ANSI implicates ANSI style function prototypes and definitions.
 *
 */
#ifndef SS_NO_ANSI

#  define SS_ANSI

#endif 

#if defined(SS_OS2) && (SS_OS2 == 32)
#  define SS_ENV_ID     SS_ENVID_O32
#  define SS_ENV_OS     SS_OS_OS2
#  define SS_ENV_OSVERS 2
#  define SS_ENV_CPU    SS_CPU_IX86
#elif defined(SS_OS2) && (SS_OS2 == 16)
#  define SS_ENV_ID     SS_ENVID_O16
#  define SS_ENV_OS     SS_OS_OS2
#  define SS_ENV_OSVERS 1
#  define SS_ENV_CPU    SS_CPU_IX86
#elif defined(SS_WIN)
#  define SS_ENV_ID     SS_ENVID_W16
#  define SS_ENV_OS     SS_OS_WIN
#  define SS_ENV_OSVERS 3
#  define SS_ENV_CPU    SS_CPU_IX86
#elif defined(SS_NT) && !defined(SS_NT64)
#  define SS_ENV_ID     (SsEnvId())
#  define SS_ENV_OS     (SsEnvOs())
#  define SS_ENV_OSVERS  (SsEnvOsVersion())
   SsEnvIdT SsEnvId(void);
   SsOsIdT SsEnvOs(void);
   int SsEnvOsVersion(void);
#  ifdef SS_ALPHA
#    define SS_ENV_CPU SS_CPU_ALPHA
#  else
#    define SS_ENV_CPU SS_CPU_IX86
#  endif

#elif defined(SS_NT64)
#  define SS_ENV_ID     SS_ENVID_WNT64
#  define SS_ENV_OS     SS_OS_WNT
#  define SS_ENV_OSVERS  (SsEnvOsVersion())
   SsEnvIdT SsEnvId(void);
   SsOsIdT SsEnvOs(void);
   int SsEnvOsVersion(void);
#  define SS_ENV_CPU    SS_CPU_AMD64
#  define SS_ENV_64BIT

#elif defined(SS_WCE)
#  define SS_ENV_OS     SS_OS_WCE
#  define SS_ENV_OSVERS 3

# ifdef TARGET_CPU_SA1100
#  define SS_ENV_CPU    SS_CPU_ARM
#  define SS_ENV_ID     SS_ENVID_WCE
# elif  TARGET_CPU_MIPSII
#  define SS_ENV_CPU    SS_CPU_MIPSII

#  ifdef WIN_CE4
#   define SS_ENV_ID     SS_ENVID_WCE4M2
#   define SS_ENV_OSVERS 4
#  else
#   define SS_ENV_ID     SS_ENVID_WCE
#   define SS_ENV_OSVERS 3
#  endif

# elif  TARGET_CPU_X86
#  define SS_ENV_CPU    SS_CPU_IX86
#  define SS_ENV_ID     SS_ENVID_WCE
#endif

#elif defined(SS_AIX) && SS_AIX >= 50
#  define SS_LARGEFILE
#  define SS_ENV_ID     SS_ENVID_A5X
#  define SS_ENV_OS     SS_OS_AIX
#  define SS_ENV_OSVERS 5
#  define SS_ENV_CPU    SS_CPU_POWERPC 
#  define AIX SS_AIX
#  if defined(SS_MT)
#    define SS_GETTIMEOFDAY_AVAILABLE
#    define SS_KERNELTHREADS
#    define SS_PTHREAD
#  endif /* SS_MT */

#elif defined(SS_AIX) && (SS_AIX >= 40 && SS_AIX < 50)
#  define SS_ENV_ID     SS_ENVID_A4X
#  define SS_ENV_OS     SS_OS_AIX
#  define SS_ENV_OSVERS 4
#  define SS_ENV_CPU    SS_CPU_POWERPC 
#  define AIX SS_AIX
#  if defined(SS_MT)
#    define SS_KERNELTHREADS
#    define SS_PTHREAD
#  endif /* SS_MT */
#elif defined(SS_AIX) && SS_AIX < 40
#  define SS_ENV_ID     SS_ENVID_A3X
#  define SS_ENV_OS     SS_OS_AIX
#  define SS_ENV_OSVERS 3
#  define SS_ENV_CPU    SS_CPU_POWERPC 
#  define AIX           SS_AIX
#elif defined(SS_VMS) && defined(SS_ALPHA)
#  define SS_ENV_ID     SS_ENVID_OVA
#  define SS_ENV_OS     SS_OS_VMS
#  define SS_ENV_OSVERS 6
#  define SS_ENV_CPU    SS_CPU_ALPHA
#elif defined(SS_VMS)
#  define SS_ENV_ID     SS_ENVID_OVV
#  define SS_ENV_OS     SS_OS_VMS
#  define SS_ENV_OSVERS 6
#  define SS_ENV_CPU    SS_CPU_VAX
#elif defined(SS_LINUX) && !defined(SS_LINUX_64BIT)
#define SS_LARGEFILE
#ifndef SS_VALGRIND /* MSG_DONTWAIT doesn't seem to work in valgrind */
#define SS_HAS_MSG_DONTWAIT
#endif
#  if defined(SS_ALPHA)
#    define SS_ENV_ID     SS_ENVID_LXA
#    define SS_ENV_CPU    SS_CPU_ALPHA
#  elif defined(SS_PPC)
#    define SS_ENV_ID     SS_ENVID_LPX
#    define SS_ENV_CPU    SS_CPU_POWERPC
# elif defined(SS_ARM)
#   define SS_ENV_ID      SS_ENVID_LSAX
#   define SS_ENV_CPU     SS_CPU_ARM
# elif defined(SS_S390)
#   define SS_ENV_ID      SS_ENVID_LS390X
#   define SS_ENV_CPU     SS_CPU_S390
# elif defined(SS_XSCALE_BE)
#    define SS_ENV_ID     SS_ENVID_LXSB
#    define SS_ENV_CPU    SS_CPU_XSCALE_BE
# elif defined(SS_SH4_BE)
#    define SS_ENV_ID     SS_ENVID_LSH4BX
#    define SS_ENV_CPU    SS_CPU_SH4_BE
# elif defined(SS_SH4_LE)
#    define SS_ENV_ID     SS_ENVID_LSH4LX
#    define SS_ENV_CPU    SS_CPU_SH4_LE
#  else 
#    define SS_ENV_ID     SS_ENVID_LUX
#    define SS_ENV_CPU    SS_CPU_IX86
/*#    define ss_int8_t     signed long long int*/
#  endif
#  define SS_ENV_OS     SS_OS_LINUX
#  define SS_ENV_OSVERS 1
#  define SS_PAGED_MEMALLOC_AVAILABLE
#  define SS_NANOSLEEP_THREADSCOPE /* suspends execution of only one thread, not all threads */
#  if defined (SS_MT)
#    define SS_PTHREAD
#    define SS_KERNELTHREADS
#    define SS_GETTIMEOFDAY_AVAILABLE
#  endif /* SS_MT */

#elif defined(SS_LINUX) && defined(SS_LINUX_64BIT)
#define SS_LARGEFILE
#ifndef SS_VALGRIND /* MSG_DONTWAIT doesn't seem to work in valgrind */
#define SS_HAS_MSG_DONTWAIT
#endif
#  define SS_ENV_ID     SS_ENVID_L2X64
#  define SS_ENV_CPU    SS_CPU_AMD64
#  define SS_ENV_OS     SS_OS_LINUX
#  define SS_ENV_OSVERS 1
#  define SS_PAGED_MEMALLOC_AVAILABLE
#  define SS_NANOSLEEP_THREADSCOPE /* suspends execution of only one thread, not all threads */
#  define SS_PTHREAD
#  define SS_KERNELTHREADS
#  define SS_GETTIMEOFDAY_AVAILABLE
#  define SS_ENV_64BIT
#elif defined(SS_SCO)
#  define SS_ENV_ID     SS_ENVID_SCX
#  define SS_ENV_OS     SS_OS_SCOUNIX
#  define SS_ENV_OSVERS 5
#  define SS_ENV_CPU    SS_CPU_IX86
#  define SCO           SS_SCO
#elif defined(SS_DOS)
#  if defined(SS_DOSX)
#    define SS_ENV_ID   SS_ENVID_D4G
#  else
#    define SS_ENV_ID   SS_ENVID_DOS
#  endif
#  define SS_ENV_OS     SS_OS_DOS
#  define SS_ENV_OSVERS 6
#  define SS_ENV_CPU    SS_CPU_IX86
#elif defined(SS_HPUX) 
#  if (SS_HPUX == 10)
#    define SS_ENV_OSVERS 10
#    define SS_ENV_ID     SS_ENVID_H0X
#  elif (SS_HPUX == 11)
#    define SS_LARGEFILE
#    define SS_ENV_OSVERS 11
#    if defined(SS_HPUX_64BIT)
#      if defined(SS_HPUX_IA64)
#        define SS_ENV_ID     SS_ENVID_HIA64BIT
#        define SS_ENV_CPU    SS_CPU_IA64
#      else
#        define SS_ENV_ID     SS_ENVID_H1X64BIT
#        define SS_ENV_CPU    SS_CPU_PA_RISC64
#      endif
#      define SS_ENV_64BIT
#    elif defined(SS_HPUX_IA64)
#	define SS_ENV_ID     SS_ENVID_HIA
#       define SS_ENV_CPU    SS_CPU_IA64_32BIT
#    else
#      define SS_ENV_ID     SS_ENVID_H1X
#      define SS_ENV_CPU    SS_CPU_PA_RISC
#    endif
#  else 
#    define SS_ENV_OSVERS 9
#    define SS_ENV_ID     SS_ENVID_H9X
#    define SS_ENV_CPU    SS_CPU_PA_RISC
#  endif  
#  define SS_ENV_OS     SS_OS_HPUX
#  if defined (SS_MT)
#    define SS_PTHREAD
#    define SS_GETTIMEOFDAY_AVAILABLE
#    if (SS_HPUX == 10)
#      define SS_PTHREAD_CMA  /* Old rev. of pthreads, was also in dix 3.2 */
#    endif
#  endif /* SS_MT */
#  define SS_PAGED_MEMALLOC_AVAILABLE
#elif defined(SS_NLM)
#  define SS_ENV_ID     (SsEnvId())
   SsEnvIdT SsEnvId(void);
#  define SS_ENV_OS     SS_OS_NW
#  define SS_ENV_OSVERS  (SsEnvOsVersion())
   int SsEnvOsVersion(void);
#  define SS_ENV_CPU    SS_CPU_IX86
#elif defined(SS_SVR4)
#  define SS_ENV_OS     SS_OS_SVR4
#  define SS_ENV_OSVERS 4
#  error  The CPU is definitely not IX86, but possibly MIPS !!!
#  define SS_ENV_CPU    SS_CPU_IX86  
#elif defined(SS_SOLARIS)
#define SS_LARGEFILE
#  if (SS_SOLARIS == 28)
#    define SS_ENV_OSVERS 28
#   if defined(SS_SOLARIS_64BIT)
#    define SS_ENV_ID     SS_ENVID_S8X64BIT
#    define SS_ENV_CPU    SS_CPU_SPARC64
#    define SS_ENV_64BIT
#   else
#    define SS_ENV_ID     SS_ENVID_S8X
#    define SS_ENV_CPU    SS_CPU_SPARC
#   endif
#  elif (SS_SOLARIS == 29)
#    define SS_ENV_OSVERS 29
#   if defined(SS_SOLARIS_64BIT)
#    define SS_ENV_ID     SS_ENVID_S9X64BIT
#    define SS_ENV_CPU    SS_CPU_SPARC64
#    define SS_ENV_64BIT
#   else
#    define SS_ENV_ID     SS_ENVID_S9X
#    define SS_ENV_CPU    SS_CPU_SPARC
#   endif
#  elif (SS_SOLARIS == 210)
#    define SS_ENV_OSVERS 210
#    if defined(SS_SPARC)
#     if defined(SS_SOLARIS_64BIT)
#      define SS_ENV_ID     SS_ENVID_S0X64BIT
#      define SS_ENV_CPU    SS_CPU_SPARC64
#      define SS_ENV_64BIT
#     else
#      define SS_ENV_ID     SS_ENVID_S0X
#      define SS_ENV_CPU    SS_CPU_SPARC
#     endif
#    else
#     if defined(SS_SOLARIS_64BIT)
#      define SS_ENV_ID     SS_ENVID_S0XI64BIT
#      define SS_ENV_CPU    SS_CPU_AMD64
#      define SS_ENV_64BIT
#     else
#      define SS_ENV_ID     SS_ENVID_S0XI
#      define SS_ENV_CPU    SS_CPU_IX86
#     endif
#    endif
#  else
#    define SS_ENV_ID     SS_ENVID_SSX
#    define SS_ENV_OSVERS 2
#  endif
#  define SS_ENV_OS     SS_OS_SOLARIS
#  define SS_NANOSLEEP_THREADSCOPE /* suspends execution of only one thread, not all threads */
#  if defined (SS_MT)
#    define SS_GETTIMEOFDAY_AVAILABLE
#    define SS_PTHREAD
#  endif /* SS_MT */
#elif defined(SS_CLIX)
#  define SS_ENV_ID     SS_ENVID_CLX
#  define SS_ENV_OS     SS_OS_CLIX
#  define SS_ENV_OSVERS 31
#  define SS_ENV_CPU    SS_CPU_CLIPPER
#elif defined(SS_DECUNIX)
#  define SS_ENV_ID     SS_ENVID_DIX
#  define SS_ENV_OS     SS_OS_DECUNIX
#  define SS_ENV_OSVERS 32
#  define SS_ENV_CPU    SS_CPU_ALPHA
#  if defined (SS_MT)
#    define SS_PTHREAD
#  endif /* SS_MT */
#elif defined(SS_IRIX)
#  define SS_ENV_ID     SS_ENVID_IRX
#  define SS_ENV_OS     SS_OS_IRIX
#  define SS_ENV_OSVERS 6
#  define SS_ENV_CPU    SS_CPU_MIPS
#  if defined (SS_MT)
#    define SS_PTHREAD
#  endif /* SS_MT */
#elif defined(SS_SUW)
#  define SS_ENV_ID     SS_ENVID_SUW
#  define SS_ENV_OS     SS_OS_SUW
#  define SS_ENV_OSVERS 7
#  define SS_ENV_CPU    SS_CPU_IX86
#elif defined(SS_VXW)
#  define SS_ENV_ID     SS_ENVID_VXW
#  define SS_ENV_OS     SS_OS_VXW
#  define SS_ENV_OSVERS 1
#  if defined(SS_PPC)
#    define SS_ENV_CPU    SS_CPU_POWERPC
#  elif defined(SS_VXSIMHP)
#    define SS_ENV_CPU    SS_CPU_VXSIMHP
#  elif defined(SS_MIPS)
#    define SS_ENV_CPU    SS_CPU_MIPS
#  elif defined(SS_VXSIMSOLARIS)
#    define SS_ENV_CPU    SS_CPU_VXSIMSOLARIS
#    define SS_ENV_ID     SS_ENVID_VSSX
#    define SS_ENV_OSVERS 541
#  else
#    define SS_ENV_CPU    SS_CPU_IX86
#  endif
#elif defined(SS_FBX)
#  if (SS_FBX >= 3)
#    define SS_LARGEFILE
#    define SS_ENV_ID     SS_ENVID_FEX
#    define SS_ENV_OSVERS 3
#    if defined(SS_FEX_64BIT)
#       define SS_GETTIMEOFDAY_AVAILABLE
#       define SS_NANOSLEEP_THREADSCOPE
#       define SS_ENV_ID     SS_ENVID_FEX64
#       define SS_ENV_OSVERS 5
#       define SS_ENV_64BIT
#    else
#       define SS_GETTIMEOFDAY_AVAILABLE
#       define SS_NANOSLEEP_THREADSCOPE
#       define SS_ENV_ID     SS_ENVID_FEX
#       define SS_ENV_OSVERS 3
#    endif
#  else
#    define SS_ENV_ID     SS_ENVID_FBX
#    define SS_ENV_OSVERS 2
#  endif
#  define SS_ENV_OS     SS_OS_FBX
#  if defined(SS_AMD64)
#       define SS_ENV_CPU    SS_CPU_AMD64
#  else
#       define SS_ENV_CPU    SS_CPU_IX86
#  endif
#elif defined(SS_CRX)
#  define SS_ENV_ID     SS_ENVID_CRX
#  define SS_ENV_OS     SS_OS_CRX
#  if defined(SS_CRX4)
#  define SS_ENV_OSVERS 4
#  else
#  define SS_ENV_OSVERS 3
#  endif
#  if defined(SS_PPC)
#    define SS_ENV_CPU    SS_CPU_POWERPC
#  elif defined(SS_SPARC)
#    define SS_ENV_CPU    SS_CPU_SPARC
#  else
#    define SS_ENV_CPU    SS_CPU_IX86
#  endif
#  if defined (SS_MT)
#    define SS_PTHREAD
#  endif /* SS_MT */
#elif defined(SS_E32)
#  define SS_ENV_ID     SS_ENVID_E32
#  define SS_ENV_OS     SS_OS_E32
#  define SS_ENV_OSVERS 3
#  define SS_ENV_CPU    SS_CPU_ARM
#elif defined(SS_PSOS)
# define SS_ENV_OSVERS  2
# define SS_ENV_ID      SS_ENVID_PSP
# define SS_ENV_CPU     SS_CPU_POWERPC
# define SS_ENV_OS      SS_OS_PSOS
#elif defined(SS_BSI)
# define SS_ENV_OSVERS  4
# define SS_ENV_ID      SS_ENVID_BSI
# define SS_ENV_CPU     SS_CPU_IX86
# define SS_ENV_OS      SS_OS_BSDI
# if defined(SS_MT)
#  define SS_PTHREAD
# endif /* SS_MT */
#elif defined(SS_OSE)
#  define SS_ENV_OS   SS_OS_OSE
#  define SS_ENV_OSVERS 4
#  if defined(SS_ARM)
#    define SS_ENV_ID   SS_ENVID_OSA
#    define SS_ENV_CPU  SS_CPU_ARM
#  elif defined(SS_SIMSOLARIS)
#    define SS_ENV_ID   SS_ENVID_OSE
#    define SS_ENV_CPU  SS_CPU_SPARC
#  elif defined(SS_PPC)
#    define SS_ENV_ID   SS_ENVID_OSE
#    define SS_ENV_CPU  SS_CPU_POWERPC
#  endif
#  if defined (SS_MT)
#    define SS_PTHREAD
#  endif /* SS_MT */
#elif defined(SS_QNX)
# define SS_ENV_OS SS_OS_QNX
# if defined(SS_PPC)
#   define SS_ENV_ID SS_ENVID_QPX
#   define SS_ENV_CPU SS_CPU_POWERPC
# else
#   define SS_ENV_ID SS_ENVID_QNX
#   define SS_ENV_CPU SS_CPU_IX86
# endif /* (SS_PPC) */
# define SS_ENV_OSVERS 6
# define SS_LARGEFILE
# define SS_NANOSLEEP_THREADSCOPE 
# define SS_GETTIMEOFDAY_AVAILABLE
/* # define SS_USE_SELECT_INSTEAD_OF_POLL */
# if defined(SS_MT)
#   define SS_PTHREAD
# endif /* SS_MT */

#else
#  error Hogihogi!
#endif

#if !defined(SS_ENV_ID)
#  error Hogihogi!
#endif

#define SS_ENV_OSNAME (SsEnvNameCurr())

char* SsEnvName(int cpu, int os, int osvers, char* outbuf, int bufsize);
char* SsEnvLicenseName(int cpu, int os, int osvers, char* outbuf, int bufsize);
const char* SsEnvCpuName(int cpu);
const char* SsEnvOsName(int os);
char* SsEnvNameCurr(void);
int   SsEnvOsversFull(int* p_major, int* p_minor);
/* char* SsEnvOsversString(void); */

/* 3-letter tokens (w32, o32, nta, etc) for Dll/lib names etc */
char*    SsEnvTokenById(SsEnvIdT envid);
SsEnvIdT SsEnvIdByToken(const char* token);
char*    SsEnvTokenCurr(void);

char*    SsHostname(void);

#if defined(SS_MT)
# define DBE_MTFLUSH /* Enables Multi-threaded flushing of db files */
# define DBE_GROUPCOMMIT /* Enables group commit in MT environments */

# if defined(SS_UNIX) && !defined(SS_IRIX)
#  if defined(SS_LINUX) /* listio has been deemed working
                           only on linux so far. */
/* #   define SS_NATIVE_LISTIO_AVAILABLE -- disabled by apl 2006-05-26 */
#  endif /* SS_LINUX */
#  if !defined(SS_DLLQMEM)
#   define SS_QMEM_THREADCTX
#  endif /* SS_DLLQMEM */

#  if !defined(SS_PTHREAD) && !defined(SS_VXWORKS)
#   error "SS_PTHREAD should be defined here!"
#  endif
# endif

#if defined(SS_VXWORKS)
#undef SS_DLLQMEM
#define SS_QMEM_THREADCTX
#undef SS_THRIDBYSTK
#define SS_PERS_HANDLES_ONLY
#if defined(SS_MT)
#define SSSYSRES_USED
#endif
#endif

#if defined(SS_CRX) && defined(SS_SPARC)
#  define SS_DLLQMEM
#  undef  SS_QMEM_THREADCTX
#  undef  SS_THRIDBYSTK
#endif

#endif

#if (defined(SS_NT) && defined(SS_MYSQL) || defined(SS_MYSQL_AC)) || defined(SS_LINUX)
#define SS_USE_INLINE
#endif

#ifdef SS_USE_INLINE

#if defined(SS_NT)
#define SS_INLINE       __inline
#elif defined(SS_SOLARIS)
#define SS_INLINE       inline
#elif defined(SS_UNIX)
#define SS_INLINE       static inline
#endif

#else /* SS_USE_INLINE */

#define SS_INLINE

#endif /* SS_USE_INLINE */

/* ---------------------------------------------------------------------
 *          Light version
 *
 */

#ifdef SS_LIGHT

#define SS_NOBACKUP         /* Database backup. */
#define SS_NOBLOB           /* Blob support. */
#define SS_NOLOCKING        /* Locking support. */
#define SS_NOLOGGING        /* Recovery log support. */
#define SS_NOMMINDEX        /* Main memory index. */
#define SS_NOREPLICATION    /* Replication. */
#define SS_NOTRXREADCHECK   /* No read checks in transactions. */
#define SS_NODDUPDATE       /* Data dictionary updates. */
#define SS_NOSQL            /* SQL support. */
#define SS_NOVIEW           /* Views. */
#define SS_NOEVENT          /* Events. */
#define SS_NOPROCEDURE      /* Procedures */
#define SS_NOSEQUENCE       /* Sequences. */
#define SS_NOERRORTEXT      /* Error texts. */
#define SS_NOCOLLATION      /* Special collations. */
#define SS_NOESTSAMPLES     /* Estimator samples for atributes. */

#endif /* SS_LIGHT */

#ifndef SS_UNICODE_DATA
# ifdef SS_DEBUG
/* #  define SS_UNICODE_DATA */
# endif /* SS_DEBUG */
#endif /* !SS_UNICODE_DATA */

#ifdef SS_VXWORKS
#include <vxWorks.h>
#define SS_SMALLSYSTEM
#endif

#ifdef SS_OSE
#define SS_SMALLSYSTEM
#endif

#ifdef SS_UNICODE_SQL
#define SS_CONNECTWITHPROLI
#define SS_PREPAREWITHPROLI
#define SS_CLIPREPAREWITHPROLI
#define SS_EXECWITHPROLI
#define SS_EXECBATCHWITHPROLI
#define SS_CLIEXECWITHPROLI
#endif /* SS_UNICODE_SQL */

#if defined(SS_PMEM) && !defined(SS_PAGED_MEMALLOC_AVAILABLE)
/* Pmem requires Paged Memalloc availability! */
#define SS_PAGED_MEMALLOC_AVAILABLE
#endif /* SS_PMEM */

/*#define SS_CATALOGSUPP*/

#ifdef SS_MME
/* These are now defaults for MME
   apl 2002-12-23 */
#define MME_SEARCH_KEEP_TVAL
#define MME_ALTERNATE_CARDINALITY
#define MME_ALTERNATE_CARDINALITY_FIX
/* This is now default for MME
   apl 2003-02-08 */
#define MME_RVAL_INDEXING
/* This is now default for MME
   apl 2003-02-13 */
#define DBE_LOCKESCALATE_OPT
/* These are now default for MME
   apl 2003-03-02 */
#define SS_TIMER_MSEC
#define DBE_LOCK_MSEC
/* This is now default for MME
   apl 2003-03-18 */
#define MME_LIKE_PREFIX
/* This is now default for MME
   apl 2003-05-21 */
#define SS_MIXED_REFERENTIAL_INTEGRITY
/* This is now default for MME
   apl 2004-01-22 */
#define MME_LOCK_FIX
/* This is now default for MME
   apl 2004-11-16 */
#define MME_CP_FIX

/* Possibility to configure the number of IO-threads per device.
   tommiv 2003-1-30 */
#if defined(SS_MT)
#ifndef SS_NTHREADS_PER_DEVQUEUE
#define SS_NTHREADS_PER_DEVQUEUE
#endif /* SS_NTHREADS_PER_DEVQUEUE */
#endif /* SS_MT */

#ifndef SS_DEBUG
#define RS_USE_MACROS
#endif

#endif /* SS_MME */

#if !defined(SS_NEW_ST)
/* checkpoint page flushing optimizations */
#  define DBE_NONBLOCKING_PAGEFLUSH
#  define DBE_NONBLOCKING_PAGEFLUSH_AVOIDCOPY
#endif /* !SS_NEW_ST */ 

#if defined(SS_PTHREAD) && !defined(SS_DEBUG) && !defined(SS_FLATSEM)
/* flattened+inlined mutex calls available for PTHREAD systems */
#define SS_FLATSEM
#endif /* SS_PTHREAD && !SS_DEBUG && !SS_FLATSEM */

#endif /* SSENV_H */
