/*************************************************************************\
**  source       * su0msgs.c
**  directory    * su
**  description  * Solid error and message texts.
**               * 
**               * Copyright (C) 2006 Solid Information Technology Ltd
\*************************************************************************/
/*
  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; only under version 2 of the License.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
  USA
*/


#include <ssstring.h>

#include "su0msgs.h"
#include "su0error.h"

/* data for searching names for return codes */
static su_rc_text_t su_rc_texts[] = {
{ SU_SUCCESS,                    SU_RCTYPE_RETCODE, "SU_SUCCESS",
  "" },

{ SU_ERR_FILE_OPEN_FAILURE,      SU_RCTYPE_ERROR,   "SU_ERR_FILE_OPEN_FAILURE",
  "File open failure" },

{ SU_ERR_FILE_WRITE_FAILURE,     SU_RCTYPE_FATAL,   "SU_ERR_FILE_WRITE_FAILURE",
  "File write failure" },

{ SU_ERR_FILE_WRITE_DISK_FULL,   SU_RCTYPE_FATAL,   "SU_ERR_FILE_WRITE_DISK_FULL",
  "File write failed, disk full" },

{ SU_ERR_FILE_WRITE_CFG_EXCEEDED, SU_RCTYPE_FATAL,  "SU_ERR_FILE_WRITE_CFG_EXCEEDED",
  "File write failed, configuration exceeded" },

{ SU_ERR_FILE_READ_FAILURE,      SU_RCTYPE_FATAL,   "SU_ERR_FILE_READ_FAILURE",
  "File read failure" },

{ SU_ERR_FILE_READ_EOF,          SU_RCTYPE_FATAL,   "SU_ERR_FILE_READ_EOF",
  "File read beyond end of file" },

{ SU_ERR_FILE_READ_ILLEGAL_ADDR, SU_RCTYPE_FATAL,   "SU_ERR_FILE_READ_ILLEGAL_ADDR",
  "File read failed, illegal file address" },

{ SU_ERR_FILE_LOCK_FAILURE,      SU_RCTYPE_FATAL,   "SU_ERR_FILE_LOCK_FAILURE",
#ifdef SS_WIN
  "File lock failure, please ensure that SHARE.EXE is installed" },
#else
  "File lock failure" },
#endif

{ SU_ERR_FILE_UNLOCK_FAILURE,    SU_RCTYPE_FATAL,   "SU_ERR_FILE_UNLOCK_FAILURE",
  "File unlock failure" },

{ SU_ERR_FILE_FREELIST_CORRUPT,  SU_RCTYPE_FATAL,   "SU_ERR_FILE_FREELIST_CORRUPT",
  "File free block list corrupted" },

{ SU_ERR_TOO_LONG_FILENAME,      SU_RCTYPE_ERROR,   "SU_ERR_TOO_LONG_FILENAME",
  "Too long file name" },
{ SU_ERR_DUPLICATE_FILENAME,     SU_RCTYPE_ERROR,   "SU_ERR_DUPLICATE_FILENAME",
  "Duplicate file name specification" },
{ SU_LI_NONE,                    SU_RCTYPE_FATAL,   "SU_LI_NONE",
  "License information not found." },
{ SU_LI_CORRUPT,                 SU_RCTYPE_FATAL,   "SU_LI_CORRUPT",
  "License file format is incompatible." },
{ SU_LI_DBTIMELIMITEXPIRED,      SU_RCTYPE_FATAL,   "SU_LI_DBTIMELIMITEXPIRED",
  "Database age limit of evaluation license expired.\n\
To continue using the database, see ORDERME.TXT for\n\
instructions on how to order a license." },
{ SU_LI_EXETIMELIMITEXPIRED,     SU_RCTYPE_FATAL,   "SU_LI_EXETIMELIMITEXPIRED",
  "Evaluation license expired.\n\
See ORDERME.TXT for instructions on how\n\
to order a license." },
{ SU_LI_CPUMISMATCH,             SU_RCTYPE_FATAL,   "SU_LI_CPUMISMATCH",
  "License is for different CPU architecture." },
{ SU_LI_OSMISMATCH,              SU_RCTYPE_FATAL,   "SU_LI_OSMISMATCH",
  "License is for different OS environment." },
{ SU_LI_OSVERSMISMATCH,          SU_RCTYPE_FATAL,   "SU_LI_OSVERSMISMATCH",
  "License is for different version of this OS." },
{ SU_LI_SOLVERSMISMATCH,         SU_RCTYPE_FATAL,   "SU_LI_SOLVERSMISMATCH",
  "License is not valid for this server version." },
{ SU_LI_CHKMISMATCH,             SU_RCTYPE_FATAL,   "SU_LI_CHKMISMATCH",
  "License information is corrupted." },
{ SU_LI_LICENSEVIOLATION,        SU_RCTYPE_FATAL,   "SU_LI_LICENSEVIOLATION",
  "Problem with Your license, please contact\n\
SOLID Information Technology ltd. immediately." },
{ SU_LI_NONETLICENSE_SS,          SU_RCTYPE_ERROR,   "SU_LI_NONETLICENSE_SS",
  "Desktop license is only for local %.256s communication, cannot use protocol %.256s for listening." },
{ SU_LI_NONETLICENSE_S,           SU_RCTYPE_ERROR,   "SU_LI_NONETLICENSE_S",
  "Desktop license is only for local communication, cannot use name '%.256s' for listening." },
{ SU_ERR_BSTREAM_BROKEN,          SU_RCTYPE_ERROR,   "SU_ERR_BSTREAM_BROKEN",
  "Internal binary stream error." },
{ SU_LI_NOTCOMPAT_S,              SU_RCTYPE_ERROR,   "SU_LI_NOTCOMPAT_S",
  "License file %.256s is not compatible with this server executable." },
{ SU_ERR_BACKUPFILENOTREMOVABLE,  SU_RCTYPE_ERROR,   "SU_ERR_BACKUPFILENOTREMOVABLE",
  "Backup directory contains file which could not be removed." },
{ SU_ERR_INVALID_SECTION,         SU_RCTYPE_ERROR,   "SU_ERR_INVALID_SECTION",
      "No such parameter section %.256s." },
{ SU_ERR_INVALID_PARAMETER,       SU_RCTYPE_ERROR,   "SU_ERR_INVALID_PARAMETER",
      "No such parameter %.256s.%.256s." },
{ SU_ERR_UNSUFF_PRIV,             SU_RCTYPE_ERROR,   "SU_ERR_UNSUFF_PRIV",
      "Not allowed to set parameter value." },
{ SU_ERR_MULTI_SET,               SU_RCTYPE_ERROR,   "SU_ERR_MULTI_SET",
      "Cannot set values to multiple parameters." },
{ SU_ERR_ILL_VALUE_TYPE,          SU_RCTYPE_ERROR,   "SU_ERR_ILL_VALUE_TYPE",
      "Illegal value type for parameter." },
{ SU_ERR_CB_ERROR,                SU_RCTYPE_ERROR,   "SU_ERR_CB_ERROR",
      "Cannot set new value for parameter %.256s.%.256s." },
{ SU_ERR_READONLY,                SU_RCTYPE_ERROR,   "SU_ERR_READONLY",
      "Parameter is read-only" },
{ SU_ERR_FILE_REMOVE_FAILURE,     SU_RCTYPE_ERROR,   "SU_ERR_FILE_REMOVE_FAILURE",
      "File remove failure." },
{ SU_ERR_FILE_ADDRESS_SPACE_EXCEEDED, SU_RCTYPE_ERROR,"SU_ERR_FILE_ADDRESS_SPACE_EXCEEDED",
      "File specification exceeds the database address space" },
{ SU_ERR_PARAM_VALUE_TOOSMALL,    SU_RCTYPE_ERROR,   "SU_ERR_PARAM_VALUE_TOOSMALL",
      "Value %.256s for parameter %.256s.%.256s is smaller than minimum value %.256s." },
{ SU_ERR_PARAM_VALUE_TOOBIG,      SU_RCTYPE_ERROR,   "SU_ERR_PARAM_VALUE_TOOBIG",
      "Value %.256s for parameter %.256s.%.256s is bigger than maximum value %.256s." },
{ SU_ERR_PARAM_VALUE_INVALID,     SU_RCTYPE_ERROR,   "SU_ERR_PARAM_VALUE_INVALID",
      "Value %.256s for parameter %.256s.%.256s is invalid." },
{ SU_ERR_PWDF_CANNOT_OPEN_S,      SU_RCTYPE_ERROR,   "SU_ERR_PWDF_CANNOT_OPEN_S",
      "Password file '%.255s' cannot be opened." },
{ SU_ERR_PWDF_NOT_FOUND_S,        SU_RCTYPE_ERROR,   "SU_ERR_PWDF_NOT_FOUND_S",
      "No password found in password file '%.255s'." }

};

/*************** Message text table for COM messages ************/
static su_rc_text_t com_msg_texts[] = {
{ COM_MSG_USER_CONNECTED_SDSS,                  SU_RCTYPE_MSG,          "COM_MSG_USER_CONNECTED_SDSS",
    "User '%.256s' connected, user id %d, machine id %.256s. %.256s\n" },
{ COM_MSG_USER_CONNECTION_TIMEOUT_SSDS,         SU_RCTYPE_MSG,          "COM_MSG_USER_CONNECTION_TIMEOUT_SSDS",
    "User '%.256s' connection%.256s timed out, user id %d, machine id %.256s. %.256s\n" },
{ COM_MSG_USER_DISCONNECTED_ABNORMALLY_SSDSS,   SU_RCTYPE_MSG,          "COM_MSG_USER_DISCONNECTED_ABNORMALLY_SSDSS",
    "User '%.256s' was disconnected abnormally%.256s, user id %d, machine id %.256s. %.256s\n" },
{ COM_MSG_USER_DISCONNECTED_SSDS,               SU_RCTYPE_MSG,          "COM_MSG_USER_DISCONNECTED_SSDS",
    "User '%.256s' disconnected%.256s, user id %d, machine id %.256s. %.256s\n" },
{ COM_MSG_ADMIN_USER_CONNECTED_SDSS,            SU_RCTYPE_MSG,          "COM_MSG_ADMIN_USER_CONNECTED_SDSS",
    "Admin user '%.256s' connected, user id %d, machine id %.256s. %.256s\n" },
{ COM_MSG_RCON_USER_CONNECTED_SDSS,             SU_RCTYPE_MSG,          "COM_MSG_RCON_USER_CONNECTED_SDSS",
    "User '%.256s' connected from a remote control, user id %d, machine id %.256s. %.256s\n" },
{ COM_MSG_TRX_IDLETIMEOUT_SDS,                  SU_RCTYPE_MSG,          "COM_MSG_TRX_IDLETIMEOUT_SDS",
    "User '%.256s' transaction idle timed out, user id %d, machine id %.256s. %.256s\n" },
{ COM_MSG_TRX_TIMEOUT_SDS,                      SU_RCTYPE_MSG,          "COM_MSG_TRX_TIMEOUT_SDS",
    "User '%.256s' transaction timed out, user id %d, machine id %.256s. %.256s\n" },
{ COM_MSG_ILLEGAL_USERNAME_SSS,                 SU_RCTYPE_MSG,          "COM_MSG_ILLEGAL_USERNAME_SSS",
    "User '%.256s' tried to connect from %.256s with an illegal username or password. %.256s\n" },
{ COM_MSG_VERSION_MISMATCH_SDDS,                SU_RCTYPE_MSG,          "COM_MSG_VERSION_MISMATCH_SDDS",
    "User '%.256s' failed to connect, version mismatch. Client version %d, server version %d. %.256s\n" },
{ COM_MSG_COLLATION_VERSION_MISMATCH_SS,        SU_RCTYPE_MSG,          "COM_MSG_COLLATION_VERSION_MISMATCH_SS",
    "User '%.256s' failed to connect, collation version mismatch. %.256s\n" },
{ COM_MSG_TOO_MANY_CLIENTS_SS,                  SU_RCTYPE_MSG,          "COM_MSG_TOO_MANY_CLIENTS_SS",
    "User '%.256s' failed to connect, there are too many connected clients. %.256s\n" },
{ COM_MSG_NEW_CONNECTIONS_ALLOWED,              SU_RCTYPE_MSG,          "COM_MSG_NEW_CONNECTIONS_ALLOWED",
    "New connections allowed.\n" },
{ COM_MSG_NEW_CONNECTIONS_NOT_ALLOWED,          SU_RCTYPE_MSG,          "COM_MSG_NEW_CONNECTIONS_NOT_ALLOWED",
    "New connections can not be allowed." },
{ COM_MSG_NO_CONNECTIONS_ALLOWED,               SU_RCTYPE_MSG,          "COM_MSG_NO_CONNECTIONS_ALLOWED",
    "No new connections allowed.\n" },
{ COM_MSG_LISTENING_STARTED_SS,                 SU_RCTYPE_MSG,          "COM_MSG_LISTENING_STARTED_SS",
    "Listening of '%.256s %.256s' started.\n" },
{ COM_MSG_LISTENING_STOPPED_SS,                 SU_RCTYPE_MSG,          "COM_MSG_LISTENING_STOPPED_SS",
    "Listening of '%.256s %.256s' stopped.\n" },
{ COM_MSG_NO_VALID_LISTENING_NAME_S,            SU_RCTYPE_MSG,          "COM_MSG_NO_VALID_LISTENING_NAME_S",
    "No valid listening name specified. Exiting from %.256s.\n" },
{ COM_MSG_SRV_IN_FATAL_STATE,                   SU_RCTYPE_MSG,          "COM_MSG_SRV_IN_FATAL_STATE",
    "Server is in fatal state, no new connections are allowed\n" },
{ COM_MSG_CANNOT_START_LISTENING_S,             SU_RCTYPE_MSG,          "COM_MSG_CANNOT_START_LISTENING_S",
    "%.250s\n" },
{ COM_MSG_UNKNOWN_CONNECTION_RECYCLING,         SU_RCTYPE_MSG,          "COM_MSG_UNKNOWN_CONNECTION_RECYCLING",
    "Unknown connection, recycling XECB." },
};


/*************** Message text table for SRV messages ************/
static su_rc_text_t srv_msg_texts[] = {
{ SRV_MSG_SHUTDOWN_BY_APP,              SU_RCTYPE_MSG,  "SRV_MSG_SHUTDOWN_BY_APP",
    "Server shut down by the application.\n" },
{ SRV_MSG_SHUTDOWN_BY_KILL,             SU_RCTYPE_MSG,  "SRV_MSG_SHUTDOWN_BY_KILL",
    "Server shut down by either ALT+F4 or kill command\n" },
{ SRV_MSG_SHUTDOWN_BY_USER_SD,          SU_RCTYPE_MSG,  "SRV_MSG_SHUTDOWN_BY_USER_SD",
"User '%.256s' issued shut down server command, user id %d.\n" },
{ SRV_MSG_SHUTDOWN_BY_UNKNOWN_USER,     SU_RCTYPE_MSG,  "SRV_MSG_SHUTDOWN_BY_UNKNOWN_USER",
    "Server shut down by unknown user (sc==NULL)\n" },
{ SRV_MSG_SHUTDOWN_ABORTED,             SU_RCTYPE_MSG,  "SRV_MSG_SHUTDOWN_ABORTED",
    "Shutdown aborted; denied by user callback.\n" },
{ SRV_MSG_SHUTDOWN_S,                   SU_RCTYPE_MSG,  "SRV_MSG_SHUTDOWN_S",
    "%.256s is shut down\n" },
{ SRV_MSG_THREAD_STILL_ACTIVE_D,        SU_RCTYPE_MSG,  "SRV_MSG_THREAD_STILL_ACTIVE_D",
    "Some thread still active, wait extra %d seconds...\n" },
{ SRV_MSG_SERVICE_INSTALLED_S,          SU_RCTYPE_MSG,  "SRV_MSG_SERVICE_INSTALLED_S",
    "Service '%.80s' installed" },
{ SRV_MSG_SERVICE_REMOVED_S,            SU_RCTYPE_MSG,  "SRV_MSG_SERVICE_REMOVED_S",
    "Service '%.80s' removed" },
{ SRV_MSG_SERVICE_INSTALL_FAILED_SDS,   SU_RCTYPE_MSG, "SRV_MSG_SERVICE_INSTALL_FAILED_SDS",
    "Install service '%.80s' failed! Error code %d (%.80s)" },
{ SRV_MSG_SERVICE_REMOVE_FAILED_SDS,    SU_RCTYPE_MSG,  "SRV_MSG_SERVICE_REMOVE_FAILED_SDS",
    "Remove service '%.80s' failed! Error code %d (%.80s)" },
{ SRV_MSG_FAILED_TO_CHANGE_DIR_S,       SU_RCTYPE_MSG,  "SRV_MSG_FAILED_TO_CHANGE_DIR_S",
    "Failed to change the current working directory to %.200s\n" },
{ SRV_MSG_WORKING_DIR_CHANGED_S,        SU_RCTYPE_MSG,  "SRV_MSG_WORKING_DIR_CHANGED_S",
    "Current working directory changed to %.512s\n" },
{ SRV_MSG_SERVICE_USAGE,                     SU_RCTYPE_MSG,  "SRV_MSG_USAGE",
    "Usage for service option: -s{start|install|remove},name,exepath,[autostart]" },
{ SRV_MSG_VERSION_S,                   SU_RCTYPE_MSG,  "SRV_MSG_VERSION_S",
    "%.256s\n" },
{ SRV_MSG_COPYRIGHT_S,                 SU_RCTYPE_MSG,  "SRV_MSG_COPYRIGHT_S",
    "%.512s\n" },
{ SRV_MSG_STARTUP_TIME_MSG_S,          SU_RCTYPE_MSG,  "SRV_MSG_STARTUP_TIME_MSG_S",
    "%.512s" },
{ SRV_MSG_FAILED_TO_START_SS,           SU_RCTYPE_MSG,  "SRV_MSG_FAILED_TO_START_SS",
    "%.200s\nFailed to start the server. Exiting from %.256s.\n" },
{ SRV_MSG_ACCESS_VIOLATION,          SU_RCTYPE_MSG,  "SRV_MSG_ACCESS_VIOLATION",
    "Causing intentionally an access violation...\n" },
{ SRV_MSG_INTENTIONAL_INTERNAL_ERROR,    SU_RCTYPE_MSG,  "SRV_MSG_INTENTIONAL_INTERNAL_ERROR",
    "Causing intentionally an internal error...\n" },
{ SRV_MSG_ACMD_ERROREXIT_D,            SU_RCTYPE_MSG,  "SRV_MSG_ACMD_ERROREXIT_D",
    "Exiting server with ADMIN COMMAND 'errorexit %d'...\n" },
{ SRV_MSG_ACMD_ASSERT,               SU_RCTYPE_MSG,  "SRV_MSG_ACMD_ASSERT",
    "Exiting server with ADMIN COMMAND 'assertexit'...\n" },
{ SRV_MSG_INVALID_LICENSE_S,           SU_RCTYPE_MSG,  "SRV_MSG_INVALID_LICENSE_S",
    "Invalid license file %.256s\n" },
{ SRV_MSG_USING_LICENSE_S,             SU_RCTYPE_MSG,  "SRV_MSG_USING_LICENSE_S",
    "Using license file %.256s\n" },
{ SRV_MSG_SIGNAL_D,                    SU_RCTYPE_MSG,  "SRV_MSG_SIGNAL_D",
    "Signal %d\n" },
{ SRV_MSG_ASSERTION,                   SU_RCTYPE_MSG,  "SRV_MSG_ASSERTION",
    "Solid Database Engine process has encountered an internal error and is unable\n\
    to continue normally. %s " },
{ SRV_MSG_ADMIN_COMMAND_S,             SU_RCTYPE_MSG,  "SRV_MSG_ADMIN_COMMAND_S",
    "Admin command: %.512s\n" },
{ SRV_MSG_ADMIN_EVENT_S,               SU_RCTYPE_MSG,  "SRV_MSG_ADMIN_EVENT_S",
    "Admin event: %.256s\n" },
{ SRV_MSG_COMMAND_LINE_S,              SU_RCTYPE_MSG,  "SRV_MSG_COMMAND_LINE_S",
    "Command line: %.256s\n" },
{ SRV_MSG_SSDEBUG_SETTING_S,           SU_RCTYPE_MSG,  "SRV_MSG_SSDEBUG_SETTING_S",
    "SS_DEBUG=%.256s\n" },
{ SRV_MSG_ASYNC_PING_TEST_COMPLETED_S, SU_RCTYPE_MSG,  "SRV_MSG_ASYNC_PING_TEST_COMPLETED_S",
    "Asynch pingtest completed successfully to '%.256s'.\n" },
{ SRV_MSG_TOO_LONG_INIFILENAME, SU_RCTYPE_MSG, "SRV_MSG_TOO_LONG_INIFILENAME",
      "Alternate inifile name is too long (>254); parameter ignored.\n" },
{ SRV_MSG_NOTSUPINBACKUPMODE,        SU_RCTYPE_MSG,    "SRV_MSG_NOTSUPINBACKUPMODE",
    "Server in backupserver mode. Operation refused" },
{ SRV_MSG_ILLUSERORPASS,             SU_RCTYPE_MSG,   "SRV_MSG_ILLUSERORPASS",
    "Connect failed, illegal user name or password" },    /* XXX */
{ SRV_MSG_THRCREATEFAIL_S,           SU_RCTYPE_MSG,   "SRV_MSG_THRCREATEFAIL_S",
    "Failed to create thread '%.80s'" },

{ SRV_MSG_PAGEDMEM_OPTION_ERROR,     SU_RCTYPE_MSG,  "SRV_MSG_PAGEDMEM_OPTION_ERROR",
    "The argument following option -x pagedmem:[client:] must be 16, 32 or 64 (default: 16)\n" },
{ SRV_MSG_TESTING_SYSTEM_PERF,       SU_RCTYPE_MSG,    "SRV_MSG_TESTING_SYSTEM_PERF",
    "Testing system performance." },
{ SRV_MSG_TESTING_WAS_SUCCESSFUL,    SU_RCTYPE_MSG,    "SRV_MSG_TESTING_WAS_SUCCESSFUL",
    "Testing was successful." },
{ SRV_MSG_TESTING_FAILED,            SU_RCTYPE_MSG,    "SRV_MSG_TESTING_FAILED",
      "Testing failed." },
{ SRV_MSG_HSBCONF_WITHNOLICENSE,     SU_RCTYPE_MSG,    "SRV_MSG_HSBCONF_WITHNOLICENSE",
      "HSB enabled server cannot operate without HotStandby license: set HotStandby.HSBEnabled to No." },
{ SRV_MSG_OPTION_SSSSS,              SU_RCTYPE_MSG,    "SRV_MSG_OPTION_S",
      "%s\n%s%s\n%.32s option is activated.\n%s" },
{ SRV_FATAL_ABORT,                   SU_RCTYPE_MSG,    "SRV_FATAL_ABORT",
      "Server emergency shutdown." },
{ SRV_FATAL_NOT_STARTED,             SU_RCTYPE_FATAL,  "SRV_FATAL_NOT_STARTED",
      "Server not started." },
{ SRV_MSG_DB_STARTED,                SU_RCTYPE_MSG,    "SRV_MSG_DB_STARTED",
     "Database started." },
{ SRV_MSG_MEMALLOC_EXCEEDED_UUU,     SU_RCTYPE_MSG,    "SRV_MSG_MEMALLOC_EXCEEDED_UUU",
      "Memory allocation size has exceeded %luMB. Current size: %lu bytes. Number of allocations: %lu.\n" },
{ SRV_MSG_MEMALLOC_DECREASED_UUU,     SU_RCTYPE_MSG,    "SRV_MSG_MEMALLOC_DECREASED_UUU",
      "Memory allocation size has fallen below %luMB. Current size: %lu bytes. Number of allocations: %lu.\n" },
{ SRV_MSG_STMTMEMTRACELIMIT_EXCEEDED_LDDDLDS, SU_RCTYPE_MSG, "SRV_MSG_STMTMEMTRACELIMIT_EXCEEDED_LDDDLDLDS",
      "Statement (id: %ld, userid: %d type: %d) has allocated %ld bytes of memory, SQL: \"%.2048s\".\n" }

};

/*************** Message text table for DBE subsys *******************/
static su_rc_text_t dbe_msg_texts[] = {
{ DBE_MSG_CREATING_NEW_DB,           SU_RCTYPE_MSG,      "DBE_MSG_CREATING_NEW_DB",
    "Creating a new database, please wait ..." },
{ DBE_MSG_DB_CONVERTED,              SU_RCTYPE_MSG,      "DBE_MSG_DB_CONVERTED",
    "Database converted successfully.\n" },
{ DBE_MSG_DB_ALREADY_EXISTS,         SU_RCTYPE_MSG,      "DBE_MSG_DB_ALREADY_EXISTS",
    "Database already exists.\n" },
{ DBE_MSG_CONVERTING_DB,             SU_RCTYPE_MSG,      "DBE_MSG_CONVERTING_DB",
    "Converting database ..." },
{ DBE_MSG_PLEASE_CONVERT_DB,         SU_RCTYPE_MSG,      "DBE_MSG_PLEASE_CONVERT_DB",
    "This database is from an older Solid version. To convert database\n\
for use with this version, start server with option -x convert.\n\
Please note that after conversion, database cannot be used with\n\
older versions of server anymore.\n" },
{ DBE_MSG_NEW_DB_NOT_CREATED,        SU_RCTYPE_MSG,      "DBE_MSG_NEW_DB_NOT_CREATED",
    "New database was not created.\n" },
{ DBE_MSG_CANNOT_CREATE_NEW_DB,      SU_RCTYPE_MSG,      "DBE_MSG_CANNOT_CREATE_NEW_DB",
    "Database does not exist. Can not create a new database, \n\
because the server is not running as a foreground process.\n\
To create a new database, start the server as a foreground \n\
process with -f option.\n" },
{ DBE_MSG_FAILED_TO_OPEN_DB_S,         SU_RCTYPE_MSG,      "DBE_MSG_FAILED_TO_OPEN_DB_S",
    "Failed to open the database. Exiting from %.256s.\n" },
{ DBE_MSG_MERGE_DENIED,              SU_RCTYPE_MSG,      "DBE_MSG_MERGE_DENIED",
    "Merge not started; denied by user callback.\n" },
{ DBE_MSG_IDLE_MERGE_STARTED_D,        SU_RCTYPE_MSG,      "DBE_MSG_IDLE_MERGE_STARTED_D",
    "Idle merge started, %ld keys to remove" },
{ DBE_MSG_MERGE_STARTED_SD,            SU_RCTYPE_MSG,      "DBE_MSG_MERGE_STARTED_SD",
    "Merge started%s, %ld keys to remove" },
{ DBE_MSG_IDLE_QUICK_MERGE_STARTED,  SU_RCTYPE_MSG,      "DBE_MSG_IDLE_QUICK_MERGE_STARTED",
    "Idle quick merge started" },
{ DBE_MSG_QUICK_MERGE_STARTED,       SU_RCTYPE_MSG,      "DBE_MSG_QUICK_MERGE_STARTED",
    "Quick merge started" },
{ DBE_MSG_MERGE_STOPPED_ALL_MERGED,  SU_RCTYPE_MSG,      "DBE_MSG_MERGE_STOPPED_ALL_MERGED",
    "Merge stopped, all keys merged" },
{ DBE_MSG_MERGE_STOPPED_N_KEYS_MERGED_D,   SU_RCTYPE_MSG,  "DBE_MSG_MERGE_STOPPED_N_KEYS_MERGED_D",
    "Merge stopped, %ld keys merged" },
{ DBE_MSG_MERGE_TASK_STARTED_D,   SU_RCTYPE_MSG,           "DBE_MSG_MERGE_TASK_STARTED_D",
    "Merge task started, %d tasks active" },
{ DBE_MSG_MERGE_USER_ENABLED,    SU_RCTYPE_MSG,            "DBE_MSG_MERGE_USER_ENABLED",
    "User merging enabled" },
{ DBE_MSG_PROCEDURE_CONVERT_ERROR_SS, SU_RCTYPE_MSG,       "DBE_MSG_PROCEDURE_CONVERT_ERROR_SS",
    "Error when converting procedures, procedure %.255s. %.512s" },
{ DBE_MSG_QUICK_MERGE_STOPPED,       SU_RCTYPE_MSG,      "DBE_MSG_QUICK_MERGE_STOPPED",
    "Quick merge stopped" },

{ DBE_MSG_CHECKING_INDEX,            SU_RCTYPE_MSG,      "DBE_MSG_CHECKING_INDEX",
    "Checking database index\n" },
{ DBE_MSG_INDEX_OK,                  SU_RCTYPE_MSG,      "DBE_MSG_INDEX_OK",
    "Database index is ok\n" },
{ DBE_MSG_CANNOT_CHECK_INDEX,        SU_RCTYPE_MSG,      "DBE_MSG_CANNOT_CHECK_INDEX",
    "Database is in backup server mode. Cannot check the index." },
{ DBE_MSG_TESTING_INDEX,             SU_RCTYPE_MSG,      "DBE_MSG_TESTING_INDEX",
    "Testing the database index." },
{ DBE_MSG_INDEX_TESTED_OK,           SU_RCTYPE_MSG,      "DBE_MSG_INDEX_TESTED_OK",
      /* FIXME: typo: succes[s]fully */
      "Database index has been tested successfully. Database index is ok." },
{ DBE_MSG_ERR_INDEX_NOT_OK,          SU_RCTYPE_MSG,      "DBE_MSG_ERR_INDEX_NOT_OK",
    "ERROR! Database index is NOT ok! Check errors from file ssdebug.log." },
{ DBE_MSG_FAILED_TO_OPEN_DB_FOR_TESTING,         SU_RCTYPE_MSG,  "DBE_MSG_FAILED_TO_OPEN_DB_FOR_TESTING",
    "SOLID Fatal Error: Failed to open the database for testing." },
{ DBE_MSG_FAILED_TO_CONNECT_DB_FOR_TESTING,      SU_RCTYPE_MSG,  "DBE_MSG_FAILED_TO_CONNECT_DB_FOR_TESTING",
    "SOLID Fatal Error: Failed to connect to the database for testing." },

{ DBE_MSG_STARTING_RECOVERY,         SU_RCTYPE_MSG,      "DBE_MSG_STARTING_RECOVERY",
    "Starting roll-forward recovery, please wait ..." },
{ DBE_MSG_RECOVERY_OF_TRXS_COMPLETED_U, SU_RCTYPE_MSG,     "DBE_MSG_RECOVERY_OF_TRXS_COMPLETED_U",
    "Recovery of %lu transactions successfully completed" },
{ DBE_MSG_RECOVERY_COMPLETED,        SU_RCTYPE_MSG,      "DBE_MSG_RECOVERY_COMPLETED",
    "Recovery successfully completed" },

{ DBE_MSG_FAILED_TO_CREATE_NEW_DB,           SU_RCTYPE_MSG,  "DBE_MSG_FAILED_TO_CREATE_NEW_DB",
    "Failed to create a new database" },
{ DBE_MSG_FAILED_TO_LOGON_TO_DB,             SU_RCTYPE_MSG,  "DBE_MSG_FAILED_TO_LOGON_TO_DB",
    "Failed to log on to the database" },
{ DBE_MSG_CONNECT_FAILS_SCRIPT_NOT_EXECUTED, SU_RCTYPE_MSG,  "DBE_MSG_CONNECT_FAILS_SCRIPT_NOT_EXECUTED",
    "Failed to connect, script not executed." },
{ DBE_MSG_FAILED_TO_OPEN_SQL_FILE,           SU_RCTYPE_MSG,  "DBE_MSG_FAILED_TO_OPEN_SQL_FILE",
    "Failed to open SQL input file" },
{ DBE_MSG_SCRIPT_FAILED_SS,                     SU_RCTYPE_MSG,  "DBE_MSG_SCRIPT_FAILED_SS",
    "Script %.80s failed:%.256s" },
{ DBE_MSG_TABLE_NOT_FOUND_S,                   SU_RCTYPE_MSG,  "DBE_MSG_TABLE_NOT_FOUND_S",
    "Table %.80s not found." },
{ DBE_MSG_CONVERTING_TABLE_S,                  SU_RCTYPE_MSG,  "DBE_MSG_CONVERTING_TABLE_S",
    "Converting table %.256s..." },
{ DBE_MSG_TABLE_CONVERTED_S,                   SU_RCTYPE_MSG,  "DBE_MSG_TABLE_CONVERTED_S",
    "Table %.256s converted" },
{ DBE_MSG_NO_NEED_TO_CONVERT_TABLE_S,          SU_RCTYPE_MSG,  "DBE_MSG_NO_NEED_TO_CONVERT_TABLE_S",
    "No need to convert table %.256s" },
{ DBE_MSG_INIT_FILESPECPROBLEM,                SU_RCTYPE_MSG,  "DBE_MSG_INIT_FILESPECPROBLEM",
    "There is a problem opening the database, \n\
because not all db files defined is the solid.ini were found. \n\
Please check the configuration. Note that only the file(s) defined \n\
with the largest FileSpec_n definition(s) should be missing.\n" },
{ DBE_MSG_USING_SPLITMERGE,                     SU_RCTYPE_MSG,  "DBE_MSG_USING_SPLITMERGE",
    "Using SplitMerge" },
{ DBE_MSG_WRITING_MMEPAGES_D,                   SU_RCTYPE_MSG,  "DBE_MSG_WRITING_MMEPAGES_D",
    "Writing MME pages to disk. Pages: %d" },
{ DBE_MSG_WRITING_MMEPAGES_FIN_D,               SU_RCTYPE_MSG,  "DBE_MSG_WRITING_MMEPAGES_FIN_D",
    "Finished writing MME pages to disk. Pages: %d" },
{ DBE_MSG_LOADING_MME_D,                        SU_RCTYPE_MSG,  "DBE_MSG_LOADING_MME_D",
    "Loading MME. Pages: %d" },
{ DBE_MSG_LOADING_MME_FIN_D,                    SU_RCTYPE_MSG,  "DBE_MSG_LOADING_MME_FIN_D",
    "Finished loading MME. Pages: %d" },
{ DBE_MSG_SHRINK_OK,                            SU_RCTYPE_MSG,  "DBE_MSG_SHRINK_OK",
    "Database file has been reorganized successfully." },
{ DBE_MSG_FAILED_TO_SHRINK,                     SU_RCTYPE_MSG,  "DBE_MSG_FAILED_TO_SHRINK",
    "ERROR! Failed to reorganize the database file! Check errors from file ssdebug.log." },
{ DBE_MSG_SHRINK_STARTED,                       SU_RCTYPE_MSG,  "DBE_MSG_SHRINK_STARTED",
  "Starting to reorganize and compact the database file." },
{ DBE_MSG_STARTING_DATABASE_RECREATE,       SU_RCTYPE_MSG,  "DBE_MSG_STARTING_DATABASE_RECREATE",
  "Starting to recreate the database (delete old database and create a new one)." },
{ DBE_MSG_DATABASE_DELETE_OK,                   SU_RCTYPE_MSG,  "DBE_MSG_DATABASE_DELETE_OK",
  "Successfully deleted database and logs" },
{ DBE_MSG_ERR_FAILED_TO_DELETE_DB,              SU_RCTYPE_MSG, "DBE_MSG_ERR_FAILED_TO_DELETE_DB",
  "Failed to delete database and/or logs, check file permissions." },
{ DBE_MSG_INIT_BROKENNETCOPY,                   SU_RCTYPE_MSG, "DBE_MSG_INIT_BROKENNETCOPY",
  "Database is a broken HSB copy or netcopy database." },
{ DBE_FATAL_CRASHAFTERCPMARK,   SU_RCTYPE_FATAL, "DBE_FATAL_CRASHAFTERCPMARK",
  "Exiting from server (FAKE_DBE_CRASHAFTERCPMARK).\n" },
{ DBE_FATAL_DBEXISTS,           SU_RCTYPE_FATAL, "DBE_FATAL_DBEXISTS",
  "Database must exist!\n" },
{ DBE_FATAL_DATERESET,          SU_RCTYPE_FATAL, "DBE_FATAL_DATERESET",
  "Database creation date is already reset!\n" },
{ DBE_FATAL_DATERESET_ONCE,     SU_RCTYPE_FATAL, "DBE_FATAL_DATERESET_ONCE",
   "Database creation time can be reset only once!\n" },
{ DBE_FATAL_ERRORTEST_SD,          SU_RCTYPE_FATAL, "DBE_FATAL_ERRORTEST_SD",
   "Error test in file %s, line %d\n" }
};


/*************** Message text table for CHECKPOINT subsys *******************/
static su_rc_text_t cp_msg_texts[] = {
{ CP_MSG_CREATION_COMPLETED,         SU_RCTYPE_MSG,      "CP_MSG_CREATION_COMPLETED",
    "Checkpoint creation completed\n" },
{ CP_MSG_CREATION_STARTED,           SU_RCTYPE_MSG,      "CP_MSG_CREATION_STARTED",
    "Checkpoint creation started\n" },
{ CP_MSG_CREATION_NOT_STARTED,       SU_RCTYPE_MSG,      "CP_MSG_CREATION_NOT_STARTED",
    "Checkpoint creation not started because shutdown is in progress\n" },
{ CP_MSG_NOT_STARTED_BECAUSE_DISABLED, SU_RCTYPE_MSG,      "CP_MSG_NOT_STARTED_BECAUSE_DISABLED",
    "Checkpoint creation not started because checkpointing is disabled\n" },
{ CP_MSG_CREATION_DENIED,            SU_RCTYPE_MSG,      "CP_MSG_CREATION_DENIED",
    "Checkpoint not started; denied by user callback.\n" },
{ CP_MSG_CREATION_START_FAILED_SS,        SU_RCTYPE_MSG,      "CP_MSG_CREATE_START_FAILED_SS",
    "Create %.256s start failed. %.256s\n" }
};

/*************** Message text table for BACKUP subsys *******************/
static su_rc_text_t backup_msg_texts[] = {
{ BACKUP_MSG_COMPLETED,              SU_RCTYPE_MSG,      "BACKUP_MSG_COMPLETED",
    "Backup completed successfully\n" },
{ BACKUP_MSG_STARTED_SS,                SU_RCTYPE_MSG,      "BACKUP_MSG_STARTED_SS",
    "Backup started to %.80s '%.80s'\n" },
{ BACKUP_MSG_START_FAILED_WITH_MSG_S,  SU_RCTYPE_MSG,      "BACKUP_MSG_START_FAILED_WITH_MSG_S",
    "Backup start failed. %.256s\n" },
{ BACKUP_MSG_ABORTED,                SU_RCTYPE_MSG,      "BACKUP_MSG_ABORTED",
    "Backup aborted.\n" },
{ BACKUP_MSG_FAILED_S,                 SU_RCTYPE_MSG,      "BACKUP_MSG_FAILED_S",
    "Backup failed. %.256s\n" },
{ BACKUP_MSG_START_DENIED,           SU_RCTYPE_MSG,      "BACKUP_MSG_START_DENIED",
    "Backup not started; denied by user callback.\n" },
{ BACKUP_MSG_NOT_SUPPORTED,          SU_RCTYPE_MSG,      "BACKUP_MSG_NOT_SUPPORTED",
    "Backup not started; Backup is not supported on diskless server.\n" },
{ BACKUP_MSG_NOT_STARTED_INDEX_FAILURE,  SU_RCTYPE_MSG,  "BACKUP_MSG_NOT_STARTED_INDEX_FAILURE",
    "Backup not started, index check failed. Errors written to file ssdebug.log.\n" },
};


/*************** Message text table for AT subsys *******************/
static su_rc_text_t at_msg_texts[] = {
{ AT_MSG_BACKUP_S,                     SU_RCTYPE_MSG,      "AT_MSG_BACKUP_S",
    "At: backup %.256s\n" },
{ AT_MSG_MAKECP,                     SU_RCTYPE_MSG,      "AT_MSG_MAKECP",
    "At: makecp\n" },
{ AT_MSG_THROWOUT_S,                   SU_RCTYPE_MSG,      "AT_MSG_THROWOUT_S",
    "At: throwout %.256s\n" },
{ AT_MSG_REPORT_S,                     SU_RCTYPE_MSG,      "AT_MSG_REPORT_S",
    "At: report %.256s\n" },
{ AT_MSG_SHUTDOWN,                   SU_RCTYPE_MSG,      "AT_MSG_SHUTDOWN",
    "At: shutdown\n" },
{ AT_MSG_SYSTEM_S,                     SU_RCTYPE_MSG,      "AT_MSG_SYSTEM_S",
    "At: system %.256s\n" },
{ AT_MSG_OPEN,                       SU_RCTYPE_MSG,      "AT_MSG_OPEN",
    "At: open\n" },
{ AT_MSG_CLOSE,                      SU_RCTYPE_MSG,      "AT_MSG_CLOSE",
    "At: close\n" },
{ AT_MSG_CLEANBGI_D,                   SU_RCTYPE_MSG,      "AT_MSG_CLEANBGI",
    "At: Deleted %d rows from SYS_BACKGROUNDJOB_INFO\n" },
{ AT_MSG_ASSERT,                     SU_RCTYPE_MSG,      "AT_MSG_ASSERT",
    "At: assert\n" },
{ AT_MSG_TIME_INCONSISTENCY,         SU_RCTYPE_MSG,      "AT_MSG_TIME_INCONSISTENCY",
    "Server noticed time inconsistency \
during at-command execution. If the system time has been changed, please restart server." },
{ AT_MSG_CMD_FAILED_S,                 SU_RCTYPE_MSG,      "AT_MSG_CMD_FAILED_S",
    "At command failed. %.256s\n" },
{ AT_MSG_ILLEGAL_IMMEDIATE_CMD_S,      SU_RCTYPE_MSG,      "AT_MSG_ILLEGAL_IMMEDIATE_CMD_S",
    "Illegal immediate at command '%.80s' ignored.\n" },
{ AT_MSG_ILLEGAL_CMD_S,                SU_RCTYPE_MSG,      "AT_MSG_ILLEGAL_CMD_S",
    "Illegal at command '%.80s' ignored.\n" }
};

/*************** Message text table for LOG subsys *******************/
static su_rc_text_t log_msg_texts[] = {
{ LOG_MSG_DISABLED,                  SU_RCTYPE_MSG,      "LOG_MSG_DISABLED",
    "Transaction logging is disabled, roll-forward recovery is not possible" },
{ LOG_MSG_USING_WRITE_MODE_D,          SU_RCTYPE_MSG,      "LOG_MSG_USING_WRITE_MODE_D",
    "Using log write mode %d" },
{ LOG_MSG_CONFLICTING_PARAMETERS,    SU_RCTYPE_MSG,      "LOG_MSG_CONFLICTING_PARAMETERS",
    "Conflicting parameters [General]BackupCopyLog=Yes and CheckpointDeleteLog=Yes" },
{ LOG_MSG_FILE_WRITE_FAILURE,        SU_RCTYPE_MSG,      "LOG_MSG_FILE_WRITE_FAILURE",
    "Log file write failure\n" },
{ LOG_MSG_CHECK_RESULTS_FROM_FILE_S,   SU_RCTYPE_MSG,      "LOG_MSG_CHECK_RESULTS_FROM_FILE_S",
    "Check results from file %.256s." },
{ LOG_MSG_CANNOT_OPEN_MSGLOG_S,        SU_RCTYPE_MSG,      "LOG_MSG_CANNOT_OPEN_MSGLOG_S",
    "Unable to open message log file '%.256s'\n" },
{ LOG_MSG_CANNOT_OPEN_TRACE_FILE_S,    SU_RCTYPE_MSG,      "LOG_MSG_CANNOT_OPEN_TRACE_FILE_S",
    "SOLID Fatal Error: Failed to open trace file %.256s." },
{ LOG_MSG_IGNORE_CORRUPTED_PART,     SU_RCTYPE_MSG,      "LOG_MSG_IGNORE_CORRUPTED_PART",
    "The tail of log was corrupt, the corrupt part was ignored." }
};

/*************** Message text table for INIFILE subsys *******************/
static su_rc_text_t inifile_msg_texts[] = {
{ INI_MSG_VALUE_NOT_MULTIPLE_OF_512_USSU,     SU_RCTYPE_MSG,      "INI_MSG_VALUE_NOT_MULTIPLE_OF_512_USSU",
    "Value %u for parameter [%.256s]%.256s is not multiple of 512, using default value %u" },
{ INI_MSG_INVALID_INDEXFILE_SPEC_SSSD,       SU_RCTYPE_MSG,      "INI_MSG_INVALID_INDEXFILE_SPEC_SSSD",
    "Value for index file specification [%.256s]%.256s is invalid, using default file %.256s and max size %ld" },
{ INI_MSG_INVALID_VALUE_IGNORE_FOLLOWING_SS,    SU_RCTYPE_MSG,  "INI_MSG_INVALID_VALUE_IGNORE_FOLLOWING_SS",
    "Value for index file specification [%.256s]%.256s is invalid, all following file specifications are ignored" },
{ INI_MSG_ILLEGAL_VALUE_USSU,                SU_RCTYPE_MSG,      "INI_MSG_ILLEGAL_VALUE_USSU",
    "Illegal value %u for parameter [%.256s]%.256s, using default value %u" },
{ INI_MSG_ILLEGAL_VALUE_SSSS,                SU_RCTYPE_MSG,      "INI_MSG_ILLEGAL_VALUE_SSSS",
    "Illegal value %.256s for parameter [%.256s]%.256s, using default value %s" },
{ INI_MSG_FAILED_TO_SAVE_FILE_S,             SU_RCTYPE_MSG,      "INI_MSG_FAILED_TO_SAVE_FILE_S",
    "Failed to save configuration file '%.256s'\n" },
{ INI_MSG_FAILED_TO_SET_MAXOPENFILES_DD,     SU_RCTYPE_MSG,      "INI_MSG_FAILED_TO_SET_MAXOPENFILES_DD",
    "Failed to set the maximum number of open files to %d, using default %d" },
{ INI_MSG_USING_CONFIG_FILE_D,               SU_RCTYPE_MSG,    "INI_MSG_USING_CONFIG_FILE_D",
    "Using configuration file %%.%ds" },
{ INI_MSG_NO_CONFIG_FILE_USING_DEFAULTS_D,   SU_RCTYPE_MSG,    "INI_MSG_NO_CONFIG_FILE_USING_DEFAULTS_D",
    "Configuration file %%.%ds not found, using defaults" },
{ INI_MSG_ILLEGAL_ISOLATION_VALUE_USSU,      SU_RCTYPE_MSG,    "INI_MSG_ILLEGAL_ISOLATION_VALUE_USSU",
    "Illegal value %u for parameter [%.256s]%.256s, using default value %u" },
{ INI_MSG_INVALID_COMMITMAXWAIT_VALUE_DSSD, SU_RCTYPE_MSG,     "INI_MSG_INVALID_COMMITMAXWAIT_VALUE_DSSD",
    "Illegal value %ld for parameter [%.256s]%.256s, using default value %ld" },
{ INI_MSG_INVALID_RELAXEDMAXDELAY_VALUE_DSSD, SU_RCTYPE_MSG,   "INI_MSG_INVALID_RELAXEDMAXDELAY_VALUE_DSSD",
    "Illegal value %ld for parameter [%.256s]%.256s, using default value %ld" },
{ INI_MSG_READTHREADMODE_TASKTHREAD_IN_HSB, SU_RCTYPE_MSG,   "INI_MSG_READTHREADMODE_TASKTHREAD_IN_HSB",
    "ReadThreadMode forced to (%ld) for parameter [%.256s]%.256s" },
{ INI_MSG_INVALID_SAFE_ACK_POLICY_VALUE_DSSD, SU_RCTYPE_MSG,   "INI_MSG_INVALID_SAFE_ACK_POLICY_VALUE_DSSD",
    "Illegal value %ld for parameter [%.256s]%.256s, using default value %ld" }
};


/*************** Message text table for HSB subsys *******************/
static su_rc_text_t hsb_msg_texts[] = {
/********* HSBROLE messages *********/
{ HSB_MSG_STARTED_AS_PRIMARY,              SU_RCTYPE_MSG,    "HSB_MSG_STARTED_AS_PRIMARY",
    "Started as a HotStandby primary\n" },
{ HSB_MSG_STARTED_AS_SECONDARY,            SU_RCTYPE_MSG,    "HSB_MSG_STARTED_AS_SECONDARY",
    "Started as a HotStandby secondary\n" },
{ HSB_MSG_STARTED_AS_SECONDARY_WITH_INFO,  SU_RCTYPE_MSG,    "HSB_MSG_STARTED_AS_SECONDARY_WITH_INFO",
    "The database was not shut down properly the last time that it was used, starting as a HotStandby secondary\n" },
{ HSB_MSG_FORCING_PRIMARY_TO_SECONDARY,    SU_RCTYPE_MSG,    "HSB_MSG_FORCING_PRIMARY_TO_SECONDARY",
    "Forcing HotStandby primary to start as a secondary\n" },
{ HSB_MSG_SWITCHED_TO_SECONDARY,           SU_RCTYPE_MSG,    "HSB_MSG_SWITCHED_TO_SECONDARY",
    "HotStandby role switched to secondary\n" },
{ HSB_MSG_SWITCHED_TO_PRIMARY,             SU_RCTYPE_MSG,    "HSB_MSG_SWITCHED_TO_PRIMARY",
    "HotStandby role switched to primary\n" },
{ HSB_MSG_SET_PRI_ALONE_OR_SWITCH_TO_SEC,  SU_RCTYPE_MSG,    "HSB_MSG_SET_PRI_ALONE_OR_SWITCH_TO_SEC",
    "Primary server must be set to PRIMARY ALONE or switched to the secondary role." },
{ HSB_MSG_PRIMARY_ALONE_ON,                SU_RCTYPE_MSG,    "HSB_MSG_PRIMARY_ALONE_ON",
    "HotStandby server set to PRIMARY ALONE." },
{ HSB_MSG_PRIMARY_ALONE_OFF,               SU_RCTYPE_MSG,    "HSB_MSG_PRIMARY_ALONE_OFF",
    "HotStandby server set to SECONDARY ALONE\n" },
{ HSB_MSG_SWITCH_TO_PRIMARY_FAILED_D,        SU_RCTYPE_MSG,    "HSB_MSG_SWITCH_TO_PRIMARY_FAILED_D",
    "HotStandby switch to primary failed, error %d\n" },
{ HSB_MSG_SWITCH_TO_SECONDARY_FAILED_D,      SU_RCTYPE_MSG,    "HSB_MSG_SWITCH_TO_SECONDARY_FAILED_D",
    "HotStandby switch to secondary failed, error %d\n" },
{ HSB_MSG_FAILED_TO_START_HSB_SD,             SU_RCTYPE_MSG,    "HSB_MSG_FAILED_TO_START_HSB_SD",
    "Failed to start HotStandby to %.80s, error %d\n" },
{ HSB_MSG_SWITCH_ROLE_TO_PRIMARY_FAILED_D,   SU_RCTYPE_MSG,    "HSB_MSG_SWITCH_ROLE_TO_PRIMARY_FAILED_D",
    "Failed to switch HotStandby role to primary, error %d\n" },
{ HSB_MSG_SWITCH_ROLE_TO_SECONDARY_FAILED_D, SU_RCTYPE_MSG,    "HSB_MSG_SWITCH_ROLE_TO_SECONDARY_FAILED_D",
    "Failed to switch HotStandby role to secondary, error %d\n" },
{ HSB_MSG_BOTH_ARE_PRIMARY_START_AS_SEC,   SU_RCTYPE_MSG,    "HSB_MSG_BOTH_ARE_PRIMARY_START_AS_SEC",
    "Both databases are primary servers, starting as a secondary\n" },
{ HSB_MSG_BOTH_ARE_PRIMARY,                SU_RCTYPE_MSG,    "HSB_MSG_BOTH_ARE_PRIMARY",
    "Both HotStandby databases are primarys\n" },
/* FIXME^: typo: primarys -> "primaries" or "primary" */
{ HSB_MSG_FAILED_TO_START_HSB_REJECTED_SD,    SU_RCTYPE_MSG,    "HSB_MSG_FAILED_TO_START_HSB_REJECTED_SD",
    "Failed to start HotStandby to %.80s, other server rejected with error %d\n" },
{ HSB_MSG_ROLE_IN_SECONDARY_SWITCHED,      SU_RCTYPE_MSG,    "HSB_MSG_ROLE_IN_SECONDARY_SWITCHED",
    "HotStandby role in secondary switched\n" },
{ HSB_MSG_SWITCHED_TO_STANDALONE,          SU_RCTYPE_MSG,    "HSB_MSG_SWITCHED_TO_STANDALONE",
    "HotStandby role switched to standalone\n" },

{ HSB_MSG_CONN_FAILED,                      SU_RCTYPE_MSG,    "HSB_MSG_CONN_FAILED",
    "HotStandby connect operation failed, %.80s\n" },

{ HSB_MSG_SET_PRIMARY_ALONE_FAILED_S,       SU_RCTYPE_MSG,    "HSB_MSG_SET_PRIMARY_ALONE_FAILED",
    "HotStandby cannot set the server to PRIMARY ALONE, %.80s." },

{ HSB_MSG_CONN_ALREADY_ACTIVE,              SU_RCTYPE_MSG,    "HSB_MSG_CONN_ALREADY_ACTIVE",
    "HotStandby connection is already active." },

{ HSB_MSG_COPY_FAILED_S,                    SU_RCTYPE_MSG,    "HSB_MSG_COPY_FAILED_S",
    "HotStandby copy failed. %.256s\n" },

{ HSB_ERR_NOTSUPP_IN_SECONDARY,             SU_RCTYPE_ERROR,    "HSB_ERR_NOTSUPP_IN_SECONDARY",
    "Server is secondary. Use primary server for this operation\n" },

{ HSB_ERR_COPY_FAILED_NOT_PRIMARY_ALONE,    SU_RCTYPE_ERROR,    "HSB_ERR_COPY_FAILED_NOT_PRIMARY_ALONE",
    "HotStandby copy or netcopy is only allowed when primary is in alone state\n" },

{ HSB_MSG_UNKNOWN_EVENT_S,                  SU_RCTYPE_MSG,    "HSB_MSG_UNKNOWN_EVENT",
    "Invalid event %s\n" },

/******** HSBLOG messages *********/
{ HSB_MSG_STARTING_CATCHUP,                SU_RCTYPE_MSG,    "HSB_MSG_STARTING_CATCHUP",
    "Starting to send HotStandby catchup data to secondary server\n" },
{ HSB_MSG_CATCHUP_COMPLETED,               SU_RCTYPE_MSG,    "HSB_MSG_CATCHUP_COMPLETED",
    "HotStandby catchup completed successfully\n" },
/* FIXME^: typo: succes[s]fully */
{ HSB_MSG_CATCHUP_ENDED_ABNORMALLY,        SU_RCTYPE_MSG,    "HSB_MSG_CATCHUP_ENDED_ABNORMALLY",
    "HotStandby catchup ended abnormally\n" },
{ HSB_MSG_CANNOT_START_CATCHUP_NOTINSYNC,  SU_RCTYPE_MSG,    "HSB_MSG_CANNOT_START_CATCHUP_NOTINSYNC",
    "HotStandby catchup can not be started. Secondary is not properly syncronized with primary, full syncronization is required\n" },
{ HSB_MSG_CATCHUP_ENDED_ABNORMALLY_STATUS_D, SU_RCTYPE_MSG,    "HSB_MSG_CATCHUP_ENDED_ABNORMALLY_STATUS_D",
    "HotStandby catchup ended abnormally, status %d\n" },
{ HSB_MSG_CATCHUP_ENDED_ABNORMALLY_ERROR_D,  SU_RCTYPE_MSG,    "HSB_MSG_CATCHUP_ENDED_ABNORMALLY_ERROR_D",
    "HotStandby catchup ended abnormally, error %d\n" },
{ HSB_MSG_CATCHUP_ENDED_WITH_COM_ERR,      SU_RCTYPE_MSG,    "HSB_MSG_CATCHUP_ENDED_WITH_COM_ERR",
    "HotStandby catchup ended abnormally due to a communication error\n" },
{ HSB_MSG_CATCHUP_ENDED_WITH_SEC_ERR_S,      SU_RCTYPE_MSG,    "HSB_MSG_CATCHUP_ENDED_WITH_SEC_ERR_S",
    "HotStandby catchup ended abnormally, secondary returned error '%.2048s'\n" },
{ HSB_MSG_LOG_SIZE_EXCEEDED_MAXIMIMUM_DD,     SU_RCTYPE_MSG,    "HSB_MSG_LOG_SIZE_EXCEEDED_MAXIMIMUM_DD",
    "HotStandby catchup size %ld greater than configured maximum size %ld, stopping HotStandby\n" },
{ HSB_MSG_FILE_ERR_IN_HSBLOG,              SU_RCTYPE_MSG,    "HSB_MSG_FILE_ERR_IN_HSBLOG",
    "File error in HotStandby catchup, stopping HotStandby\n" },
{ HSB_MSG_STARTING_TO_RECEIVE_CATCHUP,     SU_RCTYPE_MSG,    "HSB_MSG_STARTING_TO_RECEIVE_CATCHUP",
    "Starting to receive HotStandby catchup data from primary server\n" },
{ HSB_MSG_LOGCORRUPTION_NOTINSYNC,  SU_RCTYPE_MSG,    "HSB_MSG_LOGCORRUPTION_NOTINSYNC",
    "Secondary is not properly syncronized with primary due to a log file corruption. Please restart secondary and execute a HSB netcopy.\n" },

/******** HSBCONN messages *********/
{ HSB_MSG_CONN_TO_SEC_BROKEN,              SU_RCTYPE_MSG,    "HSB_MSG_CONN_TO_SEC_BROKEN",
    "Connection broken to HotStandby secondary server\n" },
{ HSB_MSG_CONN_TO_SEC_SS,                     SU_RCTYPE_MSG,    "HSB_MSG_CONN_TO_SEC_SS",
    "Connected to HotStandby %.80s %.80s\n" },
{ HSB_MSG_SEC_CONNECTED,                   SU_RCTYPE_MSG,    "HSB_MSG_SEC_CONNECTED",
    "HotStandby secondary connected\n" },
{ HSB_MSG_PRI_CONNECTED,                   SU_RCTYPE_MSG,    "HSB_MSG_PRI_CONNECTED",
    "HotStandby primary connected\n" },
{ HSB_MSG_CONN_BROKEN_WITH_OPEN_TRX,       SU_RCTYPE_MSG,    "HSB_MSG_CONN_BROKEN_WITH_OPEN_TRX",
    "Hot Standby connection broken to Secondary server with an open transaction, \
waiting for the operator to resolve transaction status. \
Primary server must be set to alone mode or switched to secondary mode.\n" },
{ HSB_MSG_PING_TIMEOUT,                    SU_RCTYPE_MSG,    "HSB_MSG_PING_TIMEOUT",
    "HotStandby ping timeout\n" },
{ HSB_MSG_CONN_TO_SEC_BROKEN_WITH_INFO_SS,    SU_RCTYPE_MSG,    "HSB_MSG_CONN_TO_SEC_BROKEN_WITH_INFO_SS",
    "Connection broken to HotStandby secondary %.80s, %.2048s\n" },
{ HSB_MSG_NOT_PROPERLY_SYNCRONIZED,        SU_RCTYPE_MSG,    "HSB_MSG_NOT_PROPERLY_SYNCRONIZED",
    "HotStandby databases are not properly syncronized\n" },
{ HSB_MSG_CONN_TO_SEC_TIMEOUT,             SU_RCTYPE_MSG,    "HSB_MSG_CONN_TO_SEC_TIMEOUT",
    "HotStandby connection to secondary timed out\n" },
{ HSB_MSG_CONN_BROKEN_WITH_ERR_S,            SU_RCTYPE_MSG,    "HSB_MSG_CONN_BROKEN_WITH_ERR_S",
    "HotStandby connection broken, %.80s\n" },
{ HSB_MSG_ERROR_MSG_S,                       SU_RCTYPE_MSG,    "HSB_MSG_ERROR_MSG_S",
    "HotStandby:%.2048s\n" },

/******** HSB NETCOPY messages *********/
{ HSB_MSG_NETCOPY_COMPLETED,               SU_RCTYPE_MSG,    "HSB_MSG_NETCOPY_COMPLETED",
    "Network backup completed.\n" },
{ HSB_MSG_STARTED_TO_RECEIVE_NETCOPY,      SU_RCTYPE_MSG,    "HSB_MSG_STARTED_TO_RECEIVE_NETCOPY",
    "Started to receive network backup.\n" },
{ HSB_MSG_DB_STARTED,                      SU_RCTYPE_MSG,    "HSB_MSG_DB_STARTED",
    "Database started using a HotStandby copy/netcopy." },
{ HSB_MSG_NETCOPY_FAILED_S,                SU_RCTYPE_MSG,    "HSB_MSG_NETCOPY_FAILED_S",
    "Network backup failed. %.256s\n" },
{ HSB_MSG_STARTED_TO_LISTEN_NETCOPY,       SU_RCTYPE_MSG,    "HSB_MSG_STARTED_TO_LISTEN_NETCOPY",
    "Database starts to listen for netcopy." },

/******** other HSB messages *********/
{ HSB_MSG_FORCING_THREADS_TO_1,            SU_RCTYPE_MSG,    "HSB_MSG_FORCING_THREADS_TO_1",
    "Hot Standby forcing threads to 1\n" },
{ HSB_MSG_CONFIGURED_BUT_LICENSE_MISSING,  SU_RCTYPE_MSG,    "HSB_MSG_CONFIGURED_BUT_LICENSE_MISSING",
    "Hot Standby replication configured but no active license found, replication not started\n" },

/******** HSBG2 messages *********/
{ SRV_RC_HSBSWITCHBEGIN,                   SU_RCTYPE_MSG,    "SRV_RC_HSBSWITCHBEGIN",
    "No role switches since the server startup\n" },
{ SRV_RC_HSBCONNECTING,                    SU_RCTYPE_MSG,    "SRV_RC_HSBCONNECTING",
    "CONNECTING" },
{ SRV_RC_HSBDISCONNECTING,                 SU_RCTYPE_MSG,    "SRV_RC_HSBDISCONNECTING",
    "DISCONNECTING" },
{ SRV_RC_HSBCATCHUP,                       SU_RCTYPE_MSG,    "SRV_RC_HSBCATCHUP",
    "CATCHUP" },
{ HSBG2_MSG_SUCC_CONNECTED,                SU_RCTYPE_MSG,    "HSBG2_MSG_SUCC_CONNECTED",
    "CONNECTED" },
{ SRV_ERR_HSBCONNBROKEN,                   SU_RCTYPE_MSG,    "SRV_ERR_HSBCONNBROKEN",
    "BROKEN" },
{ HSBG2_MSG_BAD_HOTSTANDBY_CMD,            SU_RCTYPE_MSG,    "HSBG2_MSG_BAD_HOTSTANDBY_CMD",
    "Bad Hot Standby command.\n" },
{ HSBG2_MSG_SET_STANDALONE,                     SU_RCTYPE_MSG,    "HSBG2_MSG_SET_STANDALONE",
    "HotStandby server is set to STANDALONE." },
{ HSBG2_MSG_DISCONNECT_STARTED,           SU_RCTYPE_MSG,    "HSBG2_MSG_DISCONNECT_STARTED",
    "Started the process of disconnecting the servers.\n" },
{ HSBG2_MSG_SWITCH_TO_PRIMARY_STARTED,           SU_RCTYPE_MSG,    "HSBG2_MSG_SWITCH_TO_PRIMARY_STARTED",
    "Started the process of switching the role to primary.\n" },
{ HSBG2_MSG_SWITCH_TO_SECONDARY_STARTED,           SU_RCTYPE_MSG,    "HSBG2_MSG_SWITCH_TO_SECONDARY_STARTED",
    "Started the process of switching the role to secondary.\n" },
{ HSBG2_MSG_CONNECT_STARTED,           SU_RCTYPE_MSG,    "HSBG2_MSG_CONNECT_STARTED",
    "Started the process of connecting the servers.\n" },
{ SRV_ERR_HSBNOSYNCDIR,                SU_RCTYPE_MSG,    "SRV_ERR_HSBNOSYNCDIR",
    "HotStandby copy directory not specified.\n" },
{ HSBG2_MSG_COPY_STARTED,              SU_RCTYPE_MSG,    "HSBG2_MSG_COPY_STARTED",
    "Copy started.\n" },

/* FIXME: mr 20040618: these nongeneric parameter mssages could be removed as
 * they are not used anymore (or should not be). */
{ HSBG2_MSG_PARAMETER_PRIMARYALONE_YES,SU_RCTYPE_MSG,    "HSBG2_MSG_PARAMETER_PRIMARYALONE_YES",
    "Parameter AutoPrimaryAlone is set to Yes.\n" },
{ HSBG2_MSG_PARAMETER_PRIMARYALONE_NO,  SU_RCTYPE_MSG,    "HSBG2_MSG_PARAMETER_PRIMARYALONE_NO",
    "Parameter AutoPrimaryAlone is set to No.\n" },
{ HSBG2_MSG_PARAMETER_CONNECT_S,         SU_RCTYPE_MSG,    "HSBG2_MSG_PARAMETER_CONNECT_S",
    "Parameter Connect is set to '%s'.\n" },
{ HSBG2_MSG_PARAMETER_COPYDIR_S,         SU_RCTYPE_MSG,    "HSBG2_MSG_PARAMETER_COPYDIR_S",
    "Parameter CopyDirectory is set to '%s'.\n" },
{ HSBG2_MSG_PARAMETER_PINGTIMEOUT_D,         SU_RCTYPE_MSG,    "HSBG2_MSG_PARAMETER_PINGTIMEOUT_D",
    "Parameter PingTimeout is set to %d milliseconds.\n" },
{ HSBG2_MSG_PARAMETER_CONNECTTIMEOUT_D,         SU_RCTYPE_MSG,    "HSBG2_MSG_PARAMETER_CONNECTTIMEOUT_D",
    "Parameter ConnectTimeout is set to %d.\n" },
/* end of non-generic parameter messages */

{ HSBG2_MSG_CONNECT_ACTIVE,                 SU_RCTYPE_MSG,  "HSBG2_MSG_CONNECT_ACTIVE",
    "HotStandby connection is already active.\n" },
{ HSBG2_MSG_DISCONNECT_ACTIVE,              SU_RCTYPE_MSG,  "HSBG2_MSG_DISCONNECT_ACTIVE",
    "Operation failed, disconnect is active.\n" },
{ HSBG2_MSG_CONNECTION_BROKEN,              SU_RCTYPE_MSG,       "HSBG2_MSG_CONNECTION_BROKEN",
    "HotStandby connection is already broken.\n" },
{ HSBG2_MSG_CONNECTION_ACTIVE,              SU_RCTYPE_MSG,       "HSBG2_MSG_ACTIVE",
    "Operation failed, because connection between the servers is active.\n" },
{ __NOTUSED_HSBG2_MSG_NO_NODEID,            SU_RCTYPE_MSG,       "__NOTUSED_HSBG2_MSG_NO_NODEID",
    "Hot Standby node identifier must be defined in the inifile.\n" },
{ HSBG2_MSG_ALREADY_STANDALONE,                  SU_RCTYPE_MSG,       "HSBG2_MSG_ALREADY_STANDALONE",
    "Server is already STANDALONE." },
{ HSBG2_MSG_MIGRATING,                      SU_RCTYPE_MSG,       "HSBG2_MSG_MIGRATING",
    "Hot Standby migration is active\n" },
{ HSBG2_MSG_ALREADY_PRIMARY_ALONE,          SU_RCTYPE_MSG,       "HSBG2_MSG_ALREADY_PRIMARY_ALONE",
    "Server is already set to primary alone.\n" },
{ HSBG2_MSG_ALREADY_SECONDARY_ALONE,        SU_RCTYPE_MSG,       "HSBG2_MSG_ALREADY_SECONDARY_ALONE",
    "Server is already set to secondary alone.\n" },
{ HSBG2_MSG_PARAMETER_SET_SD,               SU_RCTYPE_MSG,    "HSBG2_MSG_PARAMETER_SET_SD",
      "Parameter %s is set to %d.\n" },
{ HSBG2_MSG_PARAMETER_SET_SS,               SU_RCTYPE_MSG,    "HSBG2_MSG_PARAMETER_SET_SS",
      "Parameter %s is set to %s.\n" },
{ HSBG2_MSG_PARAMETER_SETSTR_SS,            SU_RCTYPE_MSG,    "HSBG2_MSG_PARAMETER_SETSTR_SS",
      "Parameter %s is set to '%s'.\n" },
{ HSBG2_MSG_BOTH_ARE_SECONDARY,             SU_RCTYPE_MSG,    "HSBG2_MSG_BOTH_ARE_SECONDARY",
      "Both servers are secondary.\n" },
{ HSBG2_FATAL_DOLOGSKIP_DLDLD,          SU_RCTYPE_FATAL,    "HSBG2_FATAL_DOLOGSKIP_DLDLD",
      "HotStandby:pri_dologskip:bad type %d, log pos %ld, log size %ld.\n" },
{ HSBG2_FATAL_LOGCOPY_WRI_DLDLD,        SU_RCTYPE_FATAL,    "HSBG2_FATAL_LOGCOPY_WRI_DLDLD",
      "HotStandby:pri_hsblogcopy_write:bad type %d, log pos %ld, log size %ld." },
{ HSBG2_FATAL_LOG_OPEN_S,               SU_RCTYPE_FATAL,    "HSBG2_FATAL_LOG_OPEN_S",
      "Failed to open hot standby replication log file '%.80s'.\n" },
{ HSBG2_FATAL_RESETFILE_LD,             SU_RCTYPE_FATAL,    "HSBG2_FATAL_RESETFILE_LD",
      "Failed to allocate memory for HotStandby log. Max Log size is %ld.\n" },
{ HSBG2_FATAL_BADTYPE_DLDLD,            SU_RCTYPE_FATAL,    "HSBG2_FATAL_BADTYPE_DLDLD",
      "HotStandby:solhsby:bad type %d, log pos %ld, log size %ld\n" }
};

/*************** Message text table for RPC subsys *******************/
static su_rc_text_t rpc_msg_texts[] = {
{ RPC_MSG_ILLEGAL_HSB_COUNTER_D,             SU_RCTYPE_MSG,    "RPC_MSG_ILLEGAL_HSB_COUNTER_D",
    "Hot Standby received an illegal counter data size %d from another server" },
{ RPC_MSG_ILLEGAL_REP_TYPE_PARAM_D,          SU_RCTYPE_MSG,    "RPC_MSG_ILLEGAL_REP_TYPE_PARAM_D",
    "Received an illegal replication type parameter %d" },
{ RPC_MSG_ILLEGAL_FREEARRAY_SIZE_D,          SU_RCTYPE_MSG,    "RPC_MSG_ILLEGAL_FREEARRAY_SIZE_D",
    "Received an illegal freearray size %d" },
{ RPC_MSG_ILLEGAL_ATTR_COUNT_DS,              SU_RCTYPE_MSG,    "RPC_MSG_ILLEGAL_ATTR_COUNT_DS",
    "Received an illegal attribute count %d, routine %.256s" },
{ RPC_MSG_ILLEGAL_RELOP_DS,                   SU_RCTYPE_MSG,    "RPC_MSG_ILLEGAL_RELOP_D",
    "Received an illegal relop %d, routine %.256s" },
{ RPC_MSG_ILLEGAL_TABLE_NAME_SS,              SU_RCTYPE_MSG,    "RPC_MSG_ILLEGAL_TABLE_NAME_SS",
    "Received an illegal table name '%.80s', routine %.256s" },
{ RPC_MSG_ILLEGAL_SELFLAGS_SIZE_DS,           SU_RCTYPE_MSG,    "RPC_MSG_ILLEGAL_SELFLAGS_SIZE_DS",
    "Received an illegal selflags size %d, routine %.256s" },
{ RPC_MSG_CURRENT_CURID_IN_FREEARR_D,        SU_RCTYPE_MSG,    "RPC_MSG_CURRENT_CURID_IN_FREEARR_D",
    "Current cursor id %d found from freearray" },
{ RPC_MSG_ILLEGAL_CURID_IN_FREEARR_D,        SU_RCTYPE_MSG,    "RPC_MSG_ILLEGAL_CURID_IN_FREEARR_D",
    "Illegal cursor id %d found from freearray" },
{ RPC_MSG_ILLEGAL_USERID_D,                  SU_RCTYPE_MSG,    "RPC_MSG_ILLEGAL_USERID_D",
    "Received an illegal user id %d" },
{ RPC_MSG_ILLEGAL_CONNECTID_D,               SU_RCTYPE_MSG,    "RPC_MSG_ILLEGAL_CONNECTID_D",
    "Received an illegal connect id %ld" },
{ RPC_MSG_ILLEGAL_SEQ_NUMBER_DD,              SU_RCTYPE_MSG,    "RPC_MSG_ILLEGAL_SEQ_NUMBER_DD",
    "Received an illegal sequence number %ld, expected %ld" },
{ RPC_MSG_ILLEGAL_CURID_D,                   SU_RCTYPE_MSG,    "RPC_MSG_ILLEGAL_CURID_D",
    "Received an illegal cursor id %d" },
{ RPC_MSG_ILLEGAL_CURID_WITH_INFO_DS,         SU_RCTYPE_MSG,    "RPC_MSG_ILLEGAL_CURID_WITH_INFO_DS",
    "Received an illegal cursor id %d, %.256s" },
{ RPC_MSG_ILLEGAL_ATTRID_IN_ORDERLIST_D,     SU_RCTYPE_MSG,    "RPC_MSG_ILLEGAL_ATTRID_IN_ORDERLIST_D",
    "Illegal attribute id %d in order list" },
{ RPC_MSG_ILLEGAL_ATTRID_IN_CONSTLIST_D,     SU_RCTYPE_MSG,    "RPC_MSG_ILLEGAL_ATTRID_IN_CONSTLIST_D",
    "Illegal attribute id %d in constraint list" },
{ RPC_MSG_ILLEGAL_ATTRID_IN_SELLIST_D,       SU_RCTYPE_MSG,    "RPC_MSG_ILLEGAL_ATTRID_IN_SELLIST_D",
    "Illegal attribute id %d in select list" },
{ RPC_MSG_ILLEGAL_LENGTH_PARAM_DS,            SU_RCTYPE_MSG,    "RPC_MSG_ILLEGAL_LENGTH_PARAM_DS",
    "Received an illegal length parameter %ld, routine %.256s" },
{ RPC_MSG_ILLEGAL_ATTR_NUMBER_PARAM_SD,       SU_RCTYPE_MSG,    "RPC_MSG_ILLEGAL_ATTR_NUMBER_PARAM_SD",
    "Received an illegal attribute number parameter, routine %.256s, nattrs %ld" },
{ RPC_MSG_CANNOT_SEND_UNICODE_OLD_CLIENT,  SU_RCTYPE_MSG,    "RPC_MSG_CANNOT_SEND_UNICODE_OLD_CLIENT",
    "Cannot send UNICODE string to old client version" },
{ RPC_MSG_ILLEGAL_TYPE_NUMBER_SDDD,             SU_RCTYPE_MSG,    "RPC_MSG_ILLEGAL_TYPE_NUMBER_SDDD",
    "Received an illegal type number, routine %.256s, types %d %d %d" },
{ RPC_MSG_ILLEGAL_DATE_ATTR_JAVA_CLIENT_S,   SU_RCTYPE_MSG,    "RPC_MSG_ILLEGAL_DATE_ATTR_JAVA_CLIENT_S",
    "Received an illegal date attribute from Java client, routine %.256s" },
{ RPC_MSG_ILLEGAL_ATTR_TYPE_PARAM_SD,         SU_RCTYPE_MSG,    "RPC_MSG_ILLEGAL_ATTR_TYPE_PARAM_SD",
    "Received an illegal attribute type parameter, routine %.256s, type %d" },
{ RPC_MSG_CORRUPTED_TUPLE_S,                 SU_RCTYPE_MSG,    "RPC_MSG_CORRUPTED_TUPLE_S",
    "Received a corrupted data tuple, routine %.256s, row length mismatch" },
{ RPC_MSG_ILLEGAL_SQLCUR_SYNCARRSIZE_D,      SU_RCTYPE_MSG,    "RPC_MSG_ILLEGAL_SQLCUR_SYNCARRSIZE_D",
    "Received an illegal SQL cursor sync array size %d" },
{ RPC_MSG_ILLEGAL_SQLCURID_D,                SU_RCTYPE_MSG,    "RPC_MSG_ILLEGAL_SQLCURID_D",
    "Received an illegal SQL cursor id %d in sync array" },
{ RPC_MSG_ILLEGAL_CONSOLE_INFO,            SU_RCTYPE_MSG,    "RPC_MSG_ILLEGAL_CONSOLE_INFO",
    "Illegal RPC console information" },
{ RPC_MSG_ILLEGAL_SESSION,                 SU_RCTYPE_MSG,    "RPC_MSG_ILLEGAL_SESSION",
    "Illegal RPC session" },
{ RPC_MSG_ILLEGAL_DONEARRSIZE_D,             SU_RCTYPE_MSG,    "RPC_MSG_ILLEGAL_DONEARRSIZE_D",
    "Received an illegal donearray size %d" },
{ RPC_MSG_ILLEGAL_SQL_STMTID_DS,              SU_RCTYPE_MSG,    "RPC_MSG_ILLEGAL_SQL_STMTID_DS",
    "Received an illegal SQL statement id %d, routine %.256s" },
{ RPC_MSG_ILLEGAL_SQL_STMTID_WITH_POS_DDS,     SU_RCTYPE_MSG,    "RPC_MSG_ILLEGAL_SQL_STMTID_WITH_POS_DDS",
    "Received an illegal SQL statement id %d, pos %d, routine %.256s" },
{ RPC_MSG_ILLEGAL_READ_BLOBID_DS,             SU_RCTYPE_MSG,    "RPC_MSG_ILLEGAL_READ_BLOBID_DS",
    "Received an illegal read BLOB id %d, routine %.256s" },
{ RPC_MSG_ILLEGAL_READ_BLOB_BUFSIZE_DS,       SU_RCTYPE_MSG,    "RPC_MSG_ILLEGAL_READ_BLOB_BUFSIZE_DS",
    "Received an illegal SQL read BLOB buffer size %ld, routine %.256s" },
{ RPC_MSG_BLOB_DATA_CRC_FAILED_DS,            SU_RCTYPE_MSG,    "RPC_MSG_BLOB_DATA_CRC_FAILED_DS",
    "BLOB data crc failed, block count = %d, routine %.256s" },
{ RPC_MSG_ILLEGAL_BLOBID_DS,                  SU_RCTYPE_MSG,    "RPC_MSG_ILLEGAL_BLOBID_DS",
    "Received an illegal BLOB id %d, routine %.256s" },
{ RPC_MSG_ILLEGAL_BLOB_PIECE_LENGTH_US,       SU_RCTYPE_MSG,    "RPC_MSG_ILLEGAL_BLOB_PIECE_LENGTH_US",
    "Received an illegal BLOB piece length %lu, routine %.256s" },
{ RPC_MSG_ILLEGAL_DATA_LENGTH_SU,             SU_RCTYPE_MSG,    "RPC_MSG_ILLEGAL_DATA_LENGTH_SU",
    "Received an illegal data length, routine %.256s, length %u" },
{ RPC_MSG_ILLEGAL_TUPLE_POSITION_D,          SU_RCTYPE_MSG,    "RPC_MSG_ILLEGAL_TUPLE_POSITION_D",
    "Illegal tuple position %d" },
{ RCP_MSG_PING_CLIENT_CONNECTED_S,           SU_RCTYPE_MSG,    "RPC_MSG_PING_CLIENT_CONNECTED_S",
    "Ping client from %.256s connected" },
{ RPC_MSG_PING_CLIENT_DISCONNECTED_S,        SU_RCTYPE_MSG,    "RPC_MSG_PING_CLIENT_DISCONNECTED_S",
    "Ping client from %.256s disconnected" },
{ RPC_MSG_SRV_ERROR_S,                       SU_RCTYPE_MSG,    "RPC_MSG_SRV_ERROR_S",
    "%.256s\n" }
};

/*************** Message text table for SNC subsys *******************/
static su_rc_text_t snc_msg_texts[] = {
{ SNC_MSG_START_PARALLEL_SYNC_HISTKEY_CONV,  SU_RCTYPE_MSG,  "SNC_MSG_START_PARALLEL_SYNC_HISTKEY_CONV",
    "Starting parallel sync history key conversion..." },
{ SNC_MSG_START_SYNC_HISTKEY_CONV,           SU_RCTYPE_MSG,  "SNC_MSG_START_SYNC_HISTKEY_CONV",
    "Starting sync history key conversion..." },
{ SNC_MSG_SYNC_HISTKEY_CONV_DONE,            SU_RCTYPE_MSG,  "SNC_MSG_SYNC_HISTKEY_CONV_DONE",
    "Sync history key conversion done" },
{ SNC_MSG_ERR_NOTMASTERDB,                   SU_RCTYPE_MSG,  "SNC_MSG_ERR_NOTMASTERDB",
  "Database is not a master database" }

};



/*************** Message text table for XS subsys *******************/
static su_rc_text_t xs_msg_texts[] = {
{ XS_MSG_UNABLE_TO_RESERVE_MEM_INFO_DDS,       SU_RCTYPE_MSG,    "XS_MSG_UNABLE_TO_RESERVE_MEM_INFO_DDS",
    "Unable to reserve requested %d memory blocks for external sorter. Only %d memory blocks were available.\nSQL: %.512s\n" },
{ XS_MSG_UNABLE_TO_RESERVE_MEM_DD,            SU_RCTYPE_MSG,    "XS_MSG_UNABLE_TO_RESERVE_MEM_DD",
    "Unable to reserve requested %d memory blocks for external sorter. Only %d memory blocks were available.\n" },
{ XS_FATAL_TEMPFILE_D,                        SU_RCTYPE_FATAL,  "XS_FATAL_TEMPFILE_D",
     "Failed to create a temporary file for local sorting (system errno = %d)" }

};

/*************** Message text table for FIL subsys *******************/
static su_rc_text_t fil_msg_texts[] = {
{ FIL_MSG_SSBLOCK_FAILED_SD,                SU_RCTYPE_MSG,      "FIL_MSG_SSBLOCK_FAILED_SD",
    "SsBLock failed, file '%.80s', error = %d\n" },
{ FIL_MSG_SSBLOCK_FAILED_SDD,               SU_RCTYPE_MSG,      "FIL_MSG_SSBLOCK_FAILED_SDD",
    "SsBLock failed, file '%.80s', error = %d, fd = %d\n" },
{ FIL_MSG_SSBOPENLOCAL_FAILED_SDDD,         SU_RCTYPE_MSG,      "FIL_MSG_SSBOPENLOCAL_FAILED_SDDD",
    "SsBOpenLocal failed, file '%.80s', error = %d, retries = %d, open files = %d\n" },
{ FIL_MSG_SSBOPENLOCAL_FAILED_VAX_SDDDDD,   SU_RCTYPE_MSG,      "FIL_MSG_SSBOPENLOCAL_FAILED_VAX_SDDDDD",
    "SsBOpenLocal failed, file '%.80s', error = %d, vaxc$error = %d, fab stv = %d, retries = %d, open files = %d\n" },
{ FIL_MSG_SSBOPENLOCAL_FAILED_VAX_SDDD,     SU_RCTYPE_MSG,      "FIL_MSG_SSBOPENLOCAL_FAILED_VAX_SDDD",
    "SsBOpenLocal failed, file '%.80s', error = %d, vaxc$error = %d, retries = %d\n" },
{ FIL_MSG_SSBOPENLOCAL_FAILED_DOS_SDDD,     SU_RCTYPE_MSG,      "FIL_MSG_SSBOPENLOCAL_FAILED_DOS_SDDD",
    "SsBOpenLocal failed, file '%.80s', error = %d, dos rc = %d, retries = %d\n", },
{ FIL_MSG_SSBOPENLOCAL_FAILED_DOS_SDD,      SU_RCTYPE_MSG,      "FIL_MSG_SSBOPENLOCAL_FAILED_DOS_SDD",
    "SsBOpenLocal failed, file '%.80s', error = %d, retries = %d\n" },
{ FIL_MSG_SSBOPEN_FAILED_SDD,               SU_RCTYPE_MSG,      "FIL_MSG_SSBOPEN_FAILED_SDD",
    "SsBOpen failed, file '%.80s', error = %d, retries = %d\n" },
{ FIL_MSG_FILEFLUSH_FAILED_DS,              SU_RCTYPE_MSG,      "FIL_MSG_FILEFLUSH_FAILED_DS",
    "File flush failed, error %d, file '%.80s'\n" },
{ FIL_MSG_FILEFLUSH_FAILED_VAX_DDS,         SU_RCTYPE_MSG,      "FIL_MSG_FILEFLUSH_FAILED_VAX_DDS",
    "File flush failed, error %d, vaxc$error = %d, file '%.80s'\n" },
{ FIL_MSG_FILEFLUSH_FAILED_DOS_DDS,         SU_RCTYPE_MSG,      "FIL_MSG_FILEFLUSH_FAILED_DOS_DDS",
    "File flush failed, error %d, dos rc %d, file '%.80s'\n" },
{ FIL_MSG_FILEFLUSHCLOSE_FAILED_DS,         SU_RCTYPE_MSG,      "FIL_MSG_FILEFLUSHCLOSE_FAILED_DS",
    "File flush close failed, error %d, file '%.80s'\n" },
{ FIL_MSG_FILEFLUSHOPEN_FAILED_DS,          SU_RCTYPE_MSG,      "FIL_MSG_FILEFLUSHOPEN_FAILED_DS",
    "File flush open failed, error %d, file '%.80s'\n" },
{ FIL_MSG_FILESIZEQUERY_FAILED_DSD,         SU_RCTYPE_MSG,       "FIL_MSG_FILESIZEQUERY_FAILED_DSD",
    "File size query failed, error %d, file '%.80s', retries %d\n", },
{ FIL_MSG_FILESIZEQUERYSEEK_FAILED_S,       SU_RCTYPE_MSG,       "FIL_MSG_FILESIZEQUERYSEEK_FAILED_S",
    "File size query seek failed, file '%.80s'\n" },
{ FIL_MSG_FILESIZECHANGE_FAILED_DSDD,       SU_RCTYPE_MSG,       "FIL_MSG_FILESIZECHANGE_FAILED_DSDD",
    "File size change failed, error %d, file '%.80s', newsize %ld, retries %d\n" },
{ FIL_MSG_FILESIZECHANGE_FAILED_S,          SU_RCTYPE_MSG,       "FIL_MSG_FILESIZECHANGE_FAILED_S",
    "File '%.80s' size change failed, not supported by Windows mmio\n" },
{ FIL_MSG_FILEREAD_FAILED_DSDD,             SU_RCTYPE_MSG,       "FIL_MSG_FILEREAD_FAILED_DSDD",
    "File read failed, error %d, file '%.80s', location %ld, retries %d\n" },
{ FIL_MSG_FILEREAD_FAILED_VAX_DSDDD,        SU_RCTYPE_MSG,       "FIL_MSG_FILEREAD_FAILED_VAX_DSDDD",
    "File read failed, error %d, file '%.80s', location %ld, retries %d, vaxc$error = %d\n" },
{ FIL_MSG_FILEREADPAGE_FAILED_DSDDDD,       SU_RCTYPE_MSG,       "FIL_MSG_FILEREADPAGE_FAILED_DSDDDD",
    "File paged read failed, error %d, file '%.80s', npages %d, pagesize %d, page address %ld, retries %d\n" },
{ FIL_MSG_FILEREADSEEK_FAILED_DSDD,         SU_RCTYPE_MSG,       "FIL_MSG_FILEREADSEEK_FAILED_DSDD",
    "File read seek failed, error %d, file '%.80s', location %ld, retries %d\n" },
{ FIL_MSG_FILEREADSEEK_FAILED_VAX_DSDDD,    SU_RCTYPE_MSG,       "FIL_MSG_FILEREADSEEK_FAILED_VAX_DSDDD",
    "File read seek failed, error %d, file '%.80s', location %ld, retries %d, vaxc$error = %d\n" },
{ FIL_MSG_FILEWRITE_FAILED_DSDD,            SU_RCTYPE_MSG,       "FIL_MSG_FILEWRITE_FAILED_DSDD",
    "File write failed, error %d, file '%.80s', location %ld, retries %d\n" },
{ FIL_MSG_FILEWRITE_FAILED_VAX_DSDDD,       SU_RCTYPE_MSG,       "FIL_MSG_FILEWRITE_FAILED_VAX_DSDDD",
    "File write failed, error %d, file '%.80s', location %ld, retries %d, vaxc$error = %d\n" },
{ FIL_MSG_FILEWRITEPAGE_FAILED_DSDDDD,      SU_RCTYPE_MSG,       "FIL_MSG_FILEWRITEPAGE_FAILED_DSDDDD",
    "File paged write failed, error %d, file '%.80s', npages %d, pagesize %d, page address %ld, retries %d\n" }, 
{ FIL_MSG_FILEWRITESEEK_FAILED_DSDD,        SU_RCTYPE_MSG,       "FIL_MSG_FILEWRITESEEK_FAILED_DSDD",
    "File write seek failed, error %d, file '%.80s', location %ld, retries %d\n" },
{ FIL_MSG_FILEWRITESEEK_FAILED_VAX_DSDDD,   SU_RCTYPE_MSG,       "FIL_MSG_FILEWRITESEEK_FAILED_VAX_DSDDD",
    "File write seek failed, error %d, file '%.80s', location %ld, retries %d, vaxc$error = %d\n" },
{ FIL_MSG_FILEWRITEEND_FAILED_DSD,          SU_RCTYPE_MSG,       "FIL_MSG_FILEWRITEEND_FAILED_DSD",
    "File write end failed, error %d, file '%.80s', retries %d\n" },
{ FIL_MSG_FILEWRITEEND_FAILED_VAX_DSDD,     SU_RCTYPE_MSG,       "FIL_MSG_FILEWRITEEND_FAILED_VAX_DSDD",
    "File write end failed, error %d, file '%.80s', retries %d, vaxc$error = %d\n" },
{ FIL_MSG_FILEAPPEND_FAILED_DSD,            SU_RCTYPE_MSG,       "FIL_MSG_FILEAPPEND_FAILED_DSD",
    "File append write failed, error %d, file '%.80s', retries %d\n" },
{ FIL_MSG_FILEAPPEND_FAILED_VAX_DSDD,       SU_RCTYPE_MSG,       "FIL_MSG_FILEAPPEND_FAILED_VAX_DSDD",
    "File append write failed, error %d, file '%.80s', retries %d, vaxc$error = %d\n" },
{ FIL_MSG_FILEAPPENDSEEK_FAILED_DSD,        SU_RCTYPE_MSG,       "FIL_MSG_FILEAPPENDSEEK_FAILED_DSD",
    "File append seek failed, error %d, file '%.80s', retries %d\n" },
{ FIL_MSG_FILEAPPENDSEEK_FAILED_VAX_DSDD,   SU_RCTYPE_MSG,       "FIL_MSG_FILEAPPENDSEEK_FAILED_VAX_DSDD",
    "File append seek failed, error %d, file '%.80s', retries %d, vaxc$error = %d\n" },
{ FIL_MSG_FILESEEK_FAILED_DSDD,             SU_RCTYPE_MSG,       "FIL_MSG_FILESEEK_FAILED_DSDD",
    "File seek failed, error %d, file '%.80s', location %ld, retries %d\n" },
{ FIL_MSG_FILESEEK_FAILED_DISKFULL_DSDDD,   SU_RCTYPE_MSG,       "FIL_MSG_FILESEEK_FAILED_DISKFULL_DSDDD",
    "File seek failed, disk full, error %d, file '%.80s', location %ld, new location %ld, retries %d\n" },
{ FIL_MSG_FILESEEKEND_FAILED_DSD,           SU_RCTYPE_MSG,       "FIL_MSG_FILESEEKEND_FAILED_DSD",
    "File seek end failed, error %d, file '%.80s', retries %d\n" },
{ FIL_MSG_FILESEEKTONEWSIZE_FAILED_DSD,     SU_RCTYPE_MSG,       "FIL_MSG_FILESEEKTONEWSIZE_FAILED_DSD",
    "File seek to new size failed, error %d, file '%.80s', newsize %ld\n" },
{ FIL_MSG_FILEEXPANDWRITE_FAILED_S,         SU_RCTYPE_MSG,       "FIL_MSG_FILEEXPANDWRITE_FAILED_S",
    "File expand write failed, file '%.80s'\n" },
{ FIL_MSG_FILEEXPANDSEEK_FAILED_S,          SU_RCTYPE_MSG,       "FIL_MSG_FILEEXPANDSEEK_FAILED_S",
    "File expand seek failed, file '%.80s'\n" },
{ FIL_MSG_VIRTUALALLOC_FAILED_D,            SU_RCTYPE_MSG,       "FIL_MSG_VIRTUALALLOC_FAILED_D",
    "VirtualAlloc failed, error = %d\n" }

};

/*************** Message text table for TAB subsys *******************/
static su_rc_text_t tab_msg_texts[] = {
{ TAB_MSG_BADCURSORSTATE_SD,               SU_RCTYPE_MSG,       "TAB_MSG_BADCURSORSTATE_SD",
    "Bad cursor state, function %s, state %d\n" },
{ TAB_MSG_FAKERANDOMTABLE_SS,               SU_RCTYPE_MSG,       "TAB_MSG_FAKERANDOMTABLE_S",
    "Table '%s' created as %s" }
};

/*************** Subsystems and their abbreviations *******************/
su_rc_subsysitem_t rc_subsys[SU_RC_NSUBSYS] = {
        { SU_RC_SUBSYS_SU,  "System",           FALSE, su_rc_texts, sizeof(su_rc_texts) / sizeof(su_rc_texts[0]) },
        { SU_RC_SUBSYS_DBE, "Database",         FALSE, NULL, 0 },
        { SU_RC_SUBSYS_TAB, "Table",            FALSE, NULL, 0 },
        { SU_RC_SUBSYS_SES, "Session",          FALSE, NULL, 0 },
        { SU_RC_SUBSYS_COM, "Communication",    FALSE, NULL, 0 },
        { SU_RC_SUBSYS_SRV, "Server",           FALSE, NULL, 0 },
        { SU_RC_SUBSYS_SQL, "SQL",              FALSE, NULL, 0 },
        { SU_RC_SUBSYS_SP,  "Procedure",        FALSE, NULL, 0 },
        { SU_RC_SUBSYS_SAP, "API",              FALSE, NULL, 0 },
        { SU_RC_SUBSYS_XS,  "Sorter",           FALSE, NULL, 0 },
        { SU_RC_SUBSYS_RPC, "RPC",              FALSE, NULL, 0 },
        { SU_RC_SUBSYS_SNC, "Synchronization",  FALSE, NULL, 0 },
        { SU_RC_SUBSYS_HSB, "HotStandby",       FALSE, NULL, 0 },
        { SU_RC_SUBSYS_SSA, "SSA",              FALSE, NULL, 0 },
        
        { SU_MSG_SUBSYS_COM,  "COM",            FALSE, com_msg_texts, sizeof(com_msg_texts) / sizeof(com_msg_texts[0]) },
        { SU_MSG_SUBSYS_SRV,  "SRV",            FALSE, srv_msg_texts, sizeof(srv_msg_texts) / sizeof(srv_msg_texts[0]) },
        { SU_MSG_SUBSYS_DBE,  "DBE",            FALSE, dbe_msg_texts, sizeof(dbe_msg_texts) / sizeof(dbe_msg_texts[0]) },
        { SU_MSG_SUBSYS_CP,   "CP",             FALSE, cp_msg_texts, sizeof(cp_msg_texts) / sizeof(cp_msg_texts[0]) },
        { SU_MSG_SUBSYS_BCKP, "BCKP",           FALSE, backup_msg_texts, sizeof(backup_msg_texts) / sizeof(backup_msg_texts[0]) },
        { SU_MSG_SUBSYS_AT,   "AT",             FALSE, at_msg_texts, sizeof(at_msg_texts) / sizeof(at_msg_texts[0]) },
        { SU_MSG_SUBSYS_LOG,  "LOG",            FALSE, log_msg_texts, sizeof(log_msg_texts) / sizeof(log_msg_texts[0]) },
        { SU_MSG_SUBSYS_INI,  "INI",            FALSE, inifile_msg_texts, sizeof(inifile_msg_texts) / sizeof(inifile_msg_texts[0]) },
        { SU_MSG_SUBSYS_HSB,  "HSB",            FALSE, hsb_msg_texts, sizeof(hsb_msg_texts) / sizeof(hsb_msg_texts[0]) },
        { SU_MSG_SUBSYS_RPC,  "RPC",            FALSE, rpc_msg_texts, sizeof(rpc_msg_texts) / sizeof(rpc_msg_texts[0]) },
        { SU_MSG_SUBSYS_SNC,  "SNC",            FALSE, snc_msg_texts, sizeof(snc_msg_texts) / sizeof(snc_msg_texts[0]) },
        { SU_MSG_SUBSYS_XS,   "XS",             FALSE, xs_msg_texts, sizeof(xs_msg_texts) / sizeof(xs_msg_texts[0]) },
        { SU_MSG_SUBSYS_FIL,  "FIL",            FALSE, fil_msg_texts, sizeof(fil_msg_texts) / sizeof(fil_msg_texts[0]) },
        { SU_MSG_SUBSYS_TAB,  "TAB",            FALSE, tab_msg_texts, sizeof(tab_msg_texts) / sizeof(tab_msg_texts[0]) }
};
