/*************************************************************************\
**  source       * tab0minisql.c
**  directory    * tab
**  description  * Table level MINI SQL functions.
**               * 
**               * Copyright (C) 2006 Solid Information Technology Ltd
\*************************************************************************/
/*
  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; only under version 2 of the License.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
  USA
*/

#ifdef DOCUMENTATION
**************************************************************************

Implementation:
--------------


Limitations:
-----------


Error handling:
--------------


Objects used:
------------


Preconditions:
-------------


Multithread considerations:
--------------------------


Example:
-------


**************************************************************************
#endif /* DOCUMENTATION */

#include <ssc.h>
#include <ssdebug.h>

#include <su0err.h>
#include <su0pars.h>

#include <rs0types.h>
#include <rs0ttype.h>
#include <rs0tval.h>
#include <rs0atype.h>
#include <rs0aval.h>
#include <rs0entna.h>

#include "tab1defs.h"
#include "tab0conn.h"
#include "tab0tli.h"
#include "tab0admi.h"
#include "tab0relh.h"
#include "tab1dd.h"

#include "tab0type.h"
#include "tab0minisql.h"

static tb_sqlunique_t* sql_mini_parse_primarykey(su_pars_match_t* m, uint nattrs, char* attrs[])
{
        tb_sqlunique_t* primkey = NULL;
        bool done = FALSE;
        bool found;
        char buf[64];
        int n,i;

        SS_PUSHNAME("sql_mini_parse_primarykey");
        primkey = (tb_sqlunique_t*)SsMemAlloc(sizeof(tb_sqlunique_t));

        primkey->len = nattrs;
        primkey->name = NULL;
        primkey->fields = (uint*)SsMemAlloc(primkey->len*sizeof(uint));
        i = 0;
        n = 0;

        while (!done) {
            if (!su_pars_get_id(m, buf, sizeof(buf))) {
                ss_error;
            }
            found = FALSE;
            for (i=0;i< (int)nattrs;i++) {
                if (SsStricmp(buf, attrs[i]) == 0) {
                    primkey->fields[n] = i;
                    found = TRUE;
                }
            }
            ss_assert(found);
            n++;
            ss_assert(n <= (int)nattrs);
            if (!su_pars_match_const(m, (char *)",")) {
                done = TRUE;
            }
        }
        primkey->len  = n;

        SS_POPNAME;
        return(primkey);
}

static rs_atype_t* sql_mini_prepare_atype(rs_sysi_t* cd, su_pars_match_t* m)
{
        rs_atype_t* atype;
        rs_err_t* errh;
        bool nullallowed = TRUE;
        bool islong = FALSE;
        char typename[32];
        char* pars = NULL; /* len, scale */
        uint len;
        uint scale;

        SS_PUSHNAME("sql_mini_prepare_atype");
        /* typename */
        if (su_pars_match_keyword(m, (char *)"LONG")) {
            islong = TRUE;
        }

        if (!su_pars_get_id(m, typename, sizeof(typename))) {
            ss_error;
        }
        /* len, scale */
        if (su_pars_match_const(m, (char *)"(")) {
            /* len */
            if (!su_pars_get_uint(m, &len)) {
                ss_error;
            }
            /* scale */
            scale = 0;
            if (su_pars_match_const(m, (char *)",")) {
                if (!su_pars_get_uint(m, &scale)) {
                    ss_error;
                }
            }
            if (!su_pars_match_const(m, (char *)")")) {
                ss_error;
            }
            pars = SsMemAlloc(64);
            if (scale == 0) {
                SsSprintf(pars,"%d", len);
            } else {
                SsSprintf(pars,"%d,%d", len, scale);
            }

        }
        /* nullable */
        if (su_pars_match_keyword(m, (char *)"NOT")) {
            if (su_pars_match_keyword(m, (char *)"NULL")) {
                nullallowed = FALSE;
            } else {
                ss_error;
            }
        }

        atype = rs_atype_create(cd, typename, pars, nullallowed, &errh);
        if (pars != NULL) {
            SsMemFree(pars);
        }
        SS_POPNAME;

        return(atype);
}

static void sql_mini_prepare_attrtype(rs_sysi_t* cd, su_pars_match_t* m, uint attr_c, char* name, rs_ttype_t* ttype)
{
        rs_atype_t* atype;
        rs_err_t* errh;

        SS_PUSHNAME("sql_mini_prepare_attrtype");

        atype = sql_mini_prepare_atype(cd, m);
        rs_ttype_setatype(cd, ttype, attr_c, atype);
        rs_ttype_sql_setaname(cd, ttype, attr_c, name);
        rs_atype_free(cd, atype);
        SS_POPNAME;
}

#define MINI_SQL_MAX_ATTRS      300

static bool sql_mini_prepare_createtable(
        void* cd,
        sqltrans_t* trans,
        char* sqlstr __attribute__ ((unused)),
        rs_err_t** p_errh,
        su_pars_match_t* m)
{
        char buf[64];
        char relname[64];
        bool succp = TRUE;
        rs_ttype_t* ttype;
        bool finishedp __attribute__ ((unused));
        bool done = FALSE;
#ifndef SS_MYSQL
        int n,i;
#endif
        uint                attr_c = 0;
        char**              attrs = NULL;

        tb_sqlunique_t*     primkey = NULL;
        uint                unique_c = 0;
        tb_sqlunique_t*     unique = NULL;
        uint                forkey_c = 0;
        tb_sqlforkey_t*     forkeys = NULL;
        uint*               def = NULL;
        rs_tval_t*          defvalue = NULL;
        uint                check_c = 0;
        char**              checks = NULL;
        char**              checknames = NULL;
        tb_dd_createrel_t   createtype = TB_DD_CREATEREL_USER;
        tb_dd_persistency_t persistencytype = TB_DD_PERSISTENCY_PERMANENT;
        tb_dd_store_t       storetype = TB_DD_STORE_DISK;
        tb_dd_durability_t  durability = TB_DD_DURABILITY_DEFAULT;
        char*               authid;
        char*               catalog;
        char*               extrainfo = NULL;

        finishedp = FALSE;

        SS_PUSHNAME("sql_mini_prepare_createtable");
        catalog = rs_sdefs_getcurrentdefcatalog();
        authid = (char *)RS_AVAL_SYSNAME;

        if (!su_pars_get_id(m, relname, sizeof(relname))) {
            ss_error;
        }

        if (!su_pars_match_const(m, (char *)"(")) {
            ss_error;
        }

        ttype = rs_ttype_create(cd);
        attrs = SsMemAlloc(MINI_SQL_MAX_ATTRS*sizeof(char *));
        attr_c = 0;
        while (!done) {
            if (su_pars_match_keyword(m, (char *)"PRIMARY")) {
                if (!su_pars_match_keyword(m, (char *)"KEY")) {
                    ss_error;
                }
                if (!su_pars_match_const(m, (char *)"(")) {
                    ss_error;
                }
                ss_assert(attr_c > 0);
                primkey = sql_mini_parse_primarykey(m, attr_c, attrs);
                if (!su_pars_match_const(m, (char *)")")) {
                    ss_error;
                }
                done = TRUE;
            } else {
                /* attribute name */
                if (!su_pars_get_id(m, buf, sizeof(buf))) {
                    ss_error;
                }
                attrs[attr_c] = SsMemStrdup(buf);
                sql_mini_prepare_attrtype(cd, m, attr_c, buf, ttype);
                attr_c++;
                ss_assert(attr_c < MINI_SQL_MAX_ATTRS);
                if (!su_pars_match_const(m, (char *)",")) {
                    done = TRUE;
                }
            }
        }

        if (su_pars_match_const(m, (char *)",")) {
            if (su_pars_match_keyword(m, (char *)"UNIQUE")) {
                if (!su_pars_match_const(m, (char *)"(")) {
                    ss_error;
                }
                ss_assert(attr_c > 0);
                unique = sql_mini_parse_primarykey(m, attr_c, attrs);
                if (!su_pars_match_const(m, (char *)")")) {
                    ss_error;
                }
                unique_c = 1;
            } else {
                ss_error;
            }
        }

        if (!su_pars_match_const(m, (char *)")")) {
            ss_error;
        }
        if (!su_pars_match_keyword(m, (char *)"")) {
            if (!su_pars_match_const(m, (char *)"@")) {
                ss_error;
            }
        }

        SS_PUSHNAME("sql_mini_prepare_createtable:tb_createrelation_ext");
        succp = tb_createrelation_ext(
                        cd,
                        trans,
                        relname,
                        authid,
                        catalog,
                        extrainfo,
                        ttype,
                        primkey,
                        unique_c,
                        unique,
                        forkey_c,
                        forkeys,
                        def,
                        defvalue,
                        check_c,
                        checks,
                        checknames,
                        createtype,
                        persistencytype,
                        storetype,
                        durability,
                        TB_RELMODE_SYSDEFAULT,
                        NULL,
                        p_errh);
        SS_POPNAME;
        rs_ttype_free(cd, ttype);

        if (primkey != NULL) {
            SsMemFree(primkey->fields);
            SsMemFree(primkey);
        }
        if (unique != NULL) {
            SsMemFree(unique->fields);
            SsMemFree(unique);
        }
        while (attr_c > 0) {
            attr_c--;
            SsMemFree(attrs[attr_c]);
        }
        SsMemFree(attrs);

        SS_POPNAME;
        return(succp);
}

static bool sql_mini_prepare_altertable(
        void* cd,
        sqltrans_t* trans,
        char* sqlstr __attribute__ ((unused)),
        rs_err_t** p_errh,
        su_pars_match_t* m)
{
        char colname[64];
        char authid[64];
        char relname[64];
        su_ret_t rc;
        char* catalog;
        rs_atype_t* atype;
        tb_relh_t* tbrelh;
        rs_relh_t* relh;
        rs_auth_t* auth;
        rs_ttype_t* ttype;
        su_list_t*  attrlist;

        SS_PUSHNAME("sql_mini_prepare_altertable");

        catalog = rs_sdefs_getcurrentdefcatalog();

        if (!su_pars_get_tablename(m, authid, sizeof(authid), relname, sizeof(relname))) {
            ss_error;
        }
        if (!su_pars_match_const(m, (char *)"ADD")) {
            ss_error;
        }
        if (!su_pars_match_const(m, (char *)"COLUMN")) {
            ss_error;
        }

        /* attribute name */
        if (!su_pars_get_id(m, colname, sizeof(colname))) {
            ss_error;
        }
        /* attribute type */
        atype = sql_mini_prepare_atype(cd, m);

        tbrelh = tb_relh_create(
                    cd,
                    trans,
                    relname,
                    authid,
                    catalog,
                    p_errh);
        if (tbrelh == NULL) {
            ss_derror;
            SS_POPNAME;
            return(FALSE);
        }

        relh = tb_relh_rsrelh(cd, tbrelh);
        ttype = rs_relh_ttype(cd, relh);
        auth = rs_sysi_auth(cd);

        attrlist = tb_dd_attrlist_init();

        tb_dd_attrlist_add(
            attrlist,
            colname,
            atype,
            auth,
            NULL,
            NULL,
            NULL,
            FALSE);

        rc = tb_dd_addattributelist(
                cd,
                trans,
                relh,
                attrlist,
                FALSE,
                p_errh);

        tb_dd_attrlist_done(attrlist);

        rs_atype_free(cd, atype);

        SS_POPNAME;

        return(rc == SU_SUCCESS);
}

static bool sql_mini_prepare_createindex(
        void* cd,
        sqltrans_t* trans,
        rs_relh_t* relh,
        char* sqlstr __attribute__ ((unused)),
        bool unique,
        rs_err_t** p_errh __attribute__ ((unused)),
        su_pars_match_t* m)
{
        char*               catalog;
        rs_ttype_t*         ttype = NULL;
        uint                attr_c = 0;
        char**              attrs = NULL;
        bool*               desc = NULL; /* FALSE */
        tb_dd_createrel_t   type = TB_DD_CREATEREL_USER;
        bool succp = FALSE;
        bool done = FALSE;
        char buf[64];
        char authid[64];
        char relname[64];
        char indexname[64];
        rs_err_t* errh = NULL;

        SS_PUSHNAME("sql_mini_prepare_createindex");
        catalog = rs_sdefs_getcurrentdefcatalog();

        if (!su_pars_get_id(m, indexname, sizeof(indexname))) {
            ss_error;
        }
        if (!su_pars_match_keyword(m, (char *)"ON")) {
            ss_error;
        }

        strcpy(authid, RS_AVAL_SYSNAME);
        if (!su_pars_get_tablename(m, authid, sizeof(authid), relname, sizeof(relname))) {
            ss_error;
        }

        if (relh == NULL) {
            rs_entname_t name;
            rs_entname_initbuf(&name,
                               catalog,
                               authid,
                               relname);

            relh = tb_dd_getrelh(cd, trans, &name, NULL, NULL);
        } else {
            ss_error;
        }
        ss_bassert(relh != NULL);

        if (!su_pars_match_const(m, (char *)"(")) {
            ss_error;
        }

        attrs = SsMemAlloc(30*sizeof(char *));
        desc = SsMemAlloc(30*sizeof(bool));
        attr_c = 0;
        while (!done) {
            if (!su_pars_get_id(m, buf, sizeof(buf))) {
                ss_error;
            }
            attrs[attr_c] = SsMemStrdup(buf);
            desc[attr_c] = FALSE;
            attr_c++;
            ss_assert(attr_c < 30);
            if (!su_pars_match_const(m, (char *)",")) {
                done = TRUE;
            }
        }
        if (!su_pars_match_const(m, (char *)")")) {
            ss_error;
        }
        if (!su_pars_match_keyword(m, (char *)"")) {
            if (!su_pars_match_const(m, (char *)"@")) {
                ss_error;
            }
        }

        ttype = rs_relh_ttype(cd, relh);

        succp = tb_createindex_ext(
                    cd,
                    trans,
                    indexname,
                    authid,
                    catalog,
                    relh,
                    ttype,
                    unique,
                    attr_c,
                    attrs,
                    desc,
                    type,
                    &errh);

        if (!succp) {
            if (rs_error_geterrcode(cd, errh) == E_KEYNAMEEXIST_S) {
                ss_pprintf_1(("Index on %s allready exist\n", relname));
                rs_error_free(cd, errh);
                succp = TRUE;
            }
        }

        while (attr_c > 0) {
            attr_c--;
            SsMemFree(attrs[attr_c]);
        }

        SsMemFree(attrs);
        SsMemFree(desc);
        ss_dassert(succp);

        SS_MEM_SETUNLINK(relh);
        rs_relh_done(cd, relh);

        SS_POPNAME;
        return(succp);
}

static bool sql_mini_prepare_value(su_pars_match_t* m, char* buf, uint bufsize, bool* p_nullval)
{
        bool negat = FALSE;
        char numbuf[32];

        SS_PUSHNAME("sql_mini_prepare_value");
        *p_nullval = FALSE;

        if (su_pars_get_stringliteral(m , buf, bufsize)) {
            SS_POPNAME;
            return(TRUE);
        }

        if (su_pars_match_keyword(m, (char *)"NULL")) {
            *p_nullval = FALSE;
            SS_POPNAME;
            return(TRUE);
        }

        if (!su_pars_match_const(m, (char *)"-")) {
            negat = TRUE;
        }

        if (su_pars_get_numeric(m, numbuf, sizeof(numbuf))) {
            if (negat) {
                SsSprintf(buf, "-%s", numbuf);
            } else {
                SsSprintf(buf, "%s", numbuf);
            }
            ss_dprintf_3(("Value final:%s\n", buf));
            SS_POPNAME;
            return(TRUE);
        }
        ss_error;
        SS_POPNAME;
        return(FALSE);
}

static bool sql_mini_prepare_insert(
        void* cd,
        sqltrans_t* trans,
        char* sqlstr,
        rs_err_t** p_errh __attribute__ ((unused)),
        su_pars_match_t* m)
{
        TliConnectT* tcon;
        TliCursorT* tcur;
        TliRetT trc = 0;

        int  ano;
        char* val;
        char buf[128];
        char authid[64];
        char relname[64];
        char* colname;
        char** cols = NULL;
        bool succp = TRUE;
        rs_ttype_t* ttype;
        bool done = FALSE;
        int n,i;
        uint                attr_c = 0;
        rs_relh_t*          relh;
        rs_entname_t        name;
        bool                nullval;
        char*               catalog;
        
#ifndef SS_MYSQL
        bool finishedp = FALSE;
        char**              attrs = NULL;
        tb_sqlunique_t*     primkey = NULL;
        uint                unique_c = 0;
        tb_sqlunique_t*     unique = NULL;
        uint                forkey_c = 0;
        tb_sqlforkey_t*     forkeys = NULL;
        uint*               def = NULL;
        rs_tval_t*          defvalue = NULL;
        uint                check_c = 0;
        char**              checks = NULL;
        char**              checknames = NULL;
        tb_dd_createrel_t   createtype = TB_DD_CREATEREL_USER;
        tb_dd_persistency_t persistencytype = TB_DD_PERSISTENCY_PERMANENT;
        tb_dd_store_t       storetype = TB_DD_STORE_DISK;
        tb_dd_durability_t  durability = TB_DD_DURABILITY_DEFAULT;
        char*               extrainfo = NULL;
#endif /* ! SS_MYSQL */

        SS_PUSHNAME("sql_mini_prepare_insert");
        catalog = rs_sdefs_getcurrentdefcatalog();

        strcpy(authid, RS_AVAL_SYSNAME);
        if (!su_pars_get_tablename(m, authid, sizeof(authid), relname, sizeof(relname))) {
            ss_error;
        }
        strcpy(authid, RS_AVAL_SYSNAME);

        rs_entname_initbuf(&name,
                           catalog,
                           authid,
                           relname);

        relh = tb_dd_getrelh(cd, trans, &name, NULL, NULL);
        ss_bassert(relh != NULL);
        ttype = rs_relh_ttype(cd, relh);

        if (su_pars_match_const(m, (char *)"(")) {
            cols = SsMemAlloc(30*sizeof(char*));
            n = 0;
            done = FALSE;
            while (!done) {
                if (!su_pars_get_id(m, buf, sizeof(buf))) {
                    ss_error;
                }
                cols[n] = SsMemStrdup(buf);
                n++;
                if (!su_pars_match_const(m, (char *)",")) {
                    done = TRUE;
                }
            }
            attr_c = n;
            if (!su_pars_match_const(m, (char *)")")) {
                ss_error;
            }
        }

        if (!su_pars_match_keyword(m, (char *)"VALUES")) {
            ss_error;
        }
        if (!su_pars_match_const(m, (char *)"(")) {
            ss_error;
        }
        ss_pprintf_1(("[%s]\n", sqlstr));

        tcon = TliConnectInit(cd);
        trans = TliGetTrans(tcon);

        tcur = TliCursorCreate(tcon,
                               RS_AVAL_DEFCATALOG,
                               (char *)RS_AVAL_SYSNAME,
                               relname);
        ss_assert(tcur != NULL);

        i = 0;
        ano = 0;
        done = FALSE;
        while (!done) {

            nullval = FALSE;
            succp = sql_mini_prepare_value(m, buf, sizeof(buf), &nullval);
            ss_assert(succp);

            if (cols == NULL) {
                colname = rs_ttype_sql_aname(cd, ttype, ano);
            } else {
                colname = cols[ano];
            }
            ss_pprintf_2(("%s = %s\n", colname, buf));

            val = buf;
            if (nullval) {
                TliCursorColSetNULL(tcur, colname);
            } else {
                TliCursorColClearNULL(tcur, colname);
                trc = TliCursorColStr(tcur, colname, &val);
            }
            ss_rc_assert(trc == TLI_RC_SUCC, TliCursorErrorCode(tcur));
            i++;
            trc = TliCursorInsert(tcur);
            ss_dassert(trc == TLI_RC_SUCC || TliTransIsFailed(tcon));
            if (!su_pars_match_const(m, (char *)",")) {
                done = TRUE;
            }
            ano++;
        }

        if (!su_pars_match_const(m, (char *)")")) {
            ss_error;
        }

        TliCursorFree(tcur);
        TliConnectDone(tcon);

        if (cols != NULL) {
            for (i=0;i< (int)attr_c;i++) {
                SsMemFree(cols[i]);
            }
            SsMemFree(cols);
        }

        SS_MEM_SETUNLINK(relh);
        rs_relh_done(cd, relh);

        SS_POPNAME;
        return(succp);
}

bool tb_minisql_execdirect(
        void* cd,
        sqltrans_t* sqltrans,
        rs_relh_t* relh,
        char* sqlstr,
        rs_err_t** p_errh)
{
        bool succp = FALSE;
        bool done = FALSE;
        bool finishedp = FALSE;
        su_pars_match_t m;
        bool unique = FALSE;

        SS_PUSHNAME("tb_minisql_execdirect");
        su_pars_match_init(&m, sqlstr);

        while (!done) {

            ss_pprintf_1(("sql_mini_prepare:[%.256s]\n", sqlstr));

            if (su_pars_match_keyword(&m, (char *)"CREATE")) {
                if (su_pars_match_keyword(&m, (char *)"TABLE")) {
                    succp = sql_mini_prepare_createtable(
                                cd,
                                sqltrans,
                                sqlstr,
                                p_errh,
                                &m);
                } 
                if (su_pars_match_keyword(&m, (char *)"UNIQUE")) {
                    unique = TRUE;
                }
                if (su_pars_match_keyword(&m, (char *)"INDEX")) { 
                    succp = sql_mini_prepare_createindex(
                                cd,
                                sqltrans,
                                relh,
                                sqlstr,
                                unique,
                                p_errh,
                                &m);
                } else {
                    SS_POPNAME;
                    return(succp);
                }
            } else if (su_pars_match_keyword(&m, (char *)"ALTER")) {
                if (su_pars_match_keyword(&m, (char *)"TABLE")) {
                    succp = sql_mini_prepare_altertable(
                                cd,
                                sqltrans,
                                sqlstr,
                                p_errh,
                                &m);
                } else {
                    SS_POPNAME;
                    return(succp);
                }
            } else if (su_pars_match_keyword(&m, (char *)"INSERT")) {
                if (!su_pars_match_keyword(&m, (char *)"INTO")) {
                    ss_error;
                }
                succp = sql_mini_prepare_insert(
                            cd,
                            sqltrans,
                            sqlstr,
                            p_errh,
                            &m);
            } else if (su_pars_match_keyword(&m, (char *)"COMMIT")) {
                if (!su_pars_match_keyword(&m, (char *)"WORK")) {
                    ss_error;
                }
                finishedp = FALSE;
                succp = TRUE;
                while (succp && !finishedp) {
                    succp = tb_trans_commit(cd, sqltrans, &finishedp, p_errh);
                }
                tb_trans_beginif(cd, sqltrans);
            }
            if (!su_pars_match_const(&m, (char *)"@")) {
                done = TRUE;
            }
        }
        SS_POPNAME;
        return(succp);
}
