/*************************************************************************\
**  source       * tab0srv.c
**  directory    * tab
**  description  * Server routines to table level is normal server code
**               * is not used.
**               *
**               * Copyright (C) 2006 Solid Information Technology Ltd
\*************************************************************************/
/*
  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; only under version 2 of the License.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
  USA
*/


#include <ssenv.h>

#include <ssc.h>
#include <ssdebug.h>
#include <sssem.h>
#include <ssthread.h>
#include <sstimer.h>

#include <ui0msg.h>

#include <su0usrid.h>

#include <dbe0db.h>

#include "tab1defs.h"
#include "tab0srv.h"
#include "tab0conn.h"

#ifndef SS_MYSQL
#include "../sc/sscapi.h"
#endif

#define CHK_LW(lw)  ss_dassert(SS_CHKPTR(lw) && (lw)->lw_chk == TBCHK_SRVLOCKWAIT)

#define SRV_MAXTHR  5

typedef struct {
        ss_debug(int lw_chk;)
        SsMesT*      lw_mes;
        long         lw_timeout;
        rs_sysi_t*   lw_cd;
} srv_lockwait_t;

typedef struct {
        SsMesT*         ti_mes;
        bool            ti_alive;
        bool            ti_running;
        tb_connect_t*   ti_tbcon;
} srv_thrinfo_t;

static bool          srv_shutdown = FALSE;

#ifdef SS_MYSQL

static bool          srv_mergeactive = FALSE;
static srv_thrinfo_t srv_thrinfo[SRV_MAXTHR];

static void SS_CALLBACK tb_srv_mergethread(void* param)
{
        srv_thrinfo_t* ti = param;
        SsMesRetT mesret;
        dbe_db_t* db;
        rs_sysi_t* cd;
        dbe_mergeadvance_t mergeret;

        /* Allocate thread local memory context for this thread. */
        SsQmemLocalCtxInit();

        ss_pprintf_1(("tb_srv_mergethread:start thread\n"));

        db = tb_getdb(ti->ti_tbcon);
        cd = tb_getclientdata(ti->ti_tbcon);

        while (!srv_shutdown) {
            mesret = SsMesRequest(ti->ti_mes, 10000L);
            ss_rc_dassert(mesret == SSMES_RC_SUCC || mesret == SSMES_RC_TIMEOUT, mesret);

            if (srv_mergeactive) {
                ss_pprintf_2(("tb_srv_mergethread:start merge\n"));
                ti->ti_running = TRUE;
                while (srv_mergeactive) {
                    mergeret = dbe_db_mergeadvance_ex(db, cd, 10, TRUE, NULL);
                    if (mergeret == DBE_MERGEADVANCE_END) {
                        ss_pprintf_2(("tb_srv_mergethread:DBE_MERGEADVANCE_END\n"));
                        break;
                    }
                    if (mergeret == DBE_MERGEADVANCE_PART_END) {
                        ss_pprintf_2(("tb_srv_mergethread:DBE_MERGEADVANCE_PART_END\n"));
                        SsThrSleep(100);
                    }
                    SsThrSwitch();
                }
                ti->ti_running = FALSE;
                ss_pprintf_2(("tb_srv_mergethread:stop merge\n"));
            }
        }
        ss_pprintf_1(("tb_srv_mergethread:stop\n"));

        ti->ti_alive = FALSE;

        SsThrExit();
}

static void SS_CALLBACK tb_srv_thread(void* param)
{
        srv_thrinfo_t* ti = param;
        SsMesRetT mesret;
        su_ret_t suret;
        dbe_db_t* db;
        rs_sysi_t* cd;
        int i;
        dbe_mergeadvance_t mergeret;
        SsTimeT lastmergetime = SsTime(NULL);

        /* Allocate thread local memory context for this thread. */
        SsQmemLocalCtxInit();

        ss_pprintf_1(("tb_srv_thread:start\n"));

        db = tb_getdb(ti->ti_tbcon);
        cd = tb_getclientdata(ti->ti_tbcon);

        while (!srv_shutdown) {
            bool start_merge;
            bool full_merge;

            ss_pprintf_4(("tb_srv_thread:dbe_db_cpchecklimit\n"));
            if (dbe_db_cpchecklimit(db)) {
                ss_pprintf_2(("tb_srv_thread:start checkpoint\n"));
                ti->ti_running = TRUE;
                ui_msg_message(CP_MSG_CREATION_STARTED);
                suret = tb_createcheckpoint(ti->ti_tbcon, FALSE);
                if (suret == SU_SUCCESS) {
                    ui_msg_message(CP_MSG_CREATION_COMPLETED);
                } else {
                    su_err_t* errh;
                    su_err_init(&errh, suret);
                    ui_msg_message(
                        CP_MSG_CREATION_START_FAILED_SS,
                        "checkpoint",
                        su_err_geterrstr(errh));
                    su_err_done(errh);
                }
                ss_pprintf_2(("tb_srv_thread:checkpoint done\n"));
            }

            ss_pprintf_4(("tb_srv_thread:dbe_db_mergechecklimit, mergewrites=%ld, time since last merge=%ld\n",
                dbe_db_getmergewrites(db), SsTime(NULL) - lastmergetime));
            start_merge = dbe_db_mergechecklimit(db);
            if (!start_merge
                && dbe_db_getmergewrites(db) > 0
                && lastmergetime + 10 < SsTime(NULL))
            {
                start_merge = TRUE;
                full_merge = TRUE;
            } else {
                full_merge = FALSE;
            }
            if (start_merge) {
                /* Merge will print start and stop messages internally. */
                ss_pprintf_2(("tb_srv_thread:start merge\n"));
                lastmergetime = SsTime(NULL);
                ti->ti_running = TRUE;
                if (full_merge) {
                    dbe_db_mergestart_full(cd, db);
                } else {
                    dbe_db_mergestart(cd, db);
                }
                srv_mergeactive = TRUE;
                for (i = 1; i < SRV_MAXTHR; i++) {
                    if (srv_thrinfo[i].ti_alive) {
                        SsMesSend(srv_thrinfo[i].ti_mes);
                    }
                }
                while (!srv_shutdown) {
                    mergeret = dbe_db_mergeadvance_ex(db, cd, 10, TRUE, NULL);
                    if (mergeret == DBE_MERGEADVANCE_END){
                        ss_pprintf_2(("tb_srv_thread:DBE_MERGEADVANCE_END\n"));
                        break;
                    }
                    SsThrSwitch();
                }
                srv_mergeactive = FALSE;
                for (i = 1; i < SRV_MAXTHR; i++) {
                    if (srv_thrinfo[i].ti_running) {
                        ss_pprintf_2(("tb_srv_thread:wait mergethread to stop running, i=%d\n", i));
                        SsThrSleep(100);
                    }
                }
                dbe_db_mergestop(db);
                ss_pprintf_2(("tb_srv_thread:merge done\n"));
            }

            if (ti->ti_running) {
                /* We did merge or checkpoint, do another check immediately. */
                ti->ti_running = FALSE;
            } else {
                mesret = SsMesRequest(ti->ti_mes, 1000L);
                ss_rc_dassert(mesret == SSMES_RC_SUCC || mesret == SSMES_RC_TIMEOUT, mesret);
            }
        }
        ss_pprintf_1(("tb_srv_thread:stop\n"));

        ti->ti_alive = FALSE;

        SsThrExit();
}

bool tb_srv_init(
        su_inifile_t* inifile,
        dbe_cryptoparams_t* cp)
{
        ss_pprintf_1(("tb_srv_init\n"));
        return(tb_init(inifile, cp));
}

void tb_srv_done(void)
{
        ss_pprintf_1(("tb_srv_done\n"));
        tb_done();
}

bool tb_srv_start(tb_database_t* tdb)
{
        int i;
        bool succp;
        SsThreadT* thr[SRV_MAXTHR];

        ss_pprintf_1(("tb_srv_start\n"));

        srv_shutdown = FALSE;
        srv_mergeactive = FALSE;

        for (i = 0; i < SRV_MAXTHR; i++) {
            srv_thrinfo[i].ti_mes = SsMesCreateLocal();
            srv_thrinfo[i].ti_tbcon = tb_sysconnect_init(tdb);
            ss_assert(srv_thrinfo[i].ti_tbcon != NULL);
            srv_thrinfo[i].ti_alive = FALSE;
            srv_thrinfo[i].ti_running = FALSE;
        }
        thr[0] = SsThrInitParam(
                    tb_srv_thread,
                    "tb_srv_thread",
                    1024*1024,
                    &srv_thrinfo[0]);
        for (i = 1; i < SRV_MAXTHR; i++) {
            thr[i] = SsThrInitParam(
                        tb_srv_mergethread,
                        "tb_srv_mergethread",
                        1024*1024,
                        &srv_thrinfo[i]);
        }
        for (i = 0; i < SRV_MAXTHR; i++) {
            succp = SsThrEnableBool(thr[i]);
            if (!succp) {
                break;
            }
            srv_thrinfo[i].ti_alive = TRUE;
        }
        if (!succp) {
            tb_srv_stop();
        }
        for (i = 0; i < SRV_MAXTHR; i++) {
            SsThrDone(thr[i]);
        }

        SsTimerGlobalInit();

        return(succp);
}

void tb_srv_stop(void)
{
        SsTimeT now;
        int nalive;
        int i;

        ss_pprintf_1(("tb_srv_stop\n"));

        srv_mergeactive = FALSE;
        srv_shutdown = TRUE;

        now = SsTime(NULL);
        do {
            nalive = 0;
            for (i = 0; i < SRV_MAXTHR; i++) {
                if (srv_thrinfo[i].ti_alive) {
                    SsMesSend(srv_thrinfo[i].ti_mes);
                    nalive++;
                }
            }
            if (nalive > 0) {
                ss_pprintf_2(("tb_srv_stop:wait, nalive=%d\n", nalive));
                SsThrSleep(100);
            }
            /* AKARYAKIN 19-JAN-2007: Disable shutdown timeout
             because possible checkpoint operation cannot be 
             cancelled and we have to wait for it to complete.
             Bugzilla bug #516 (int).
            */
        } while (nalive > 0 /* && SsTime(NULL) < now + 10 */);
        for (i = 0; i < SRV_MAXTHR; i++) {
            tb_sysconnect_done(srv_thrinfo[i].ti_tbcon);
            SsMesFree(srv_thrinfo[i].ti_mes);
        }
        SsTimerGlobalDone();
        ss_pprintf_2(("tb_srv_stop:return\n"));
}

#else /* SS_MYSQL */

static int solid_started = 0;
static SsSemT* solid_mutex = NULL;
static ssc_serverhandle_t solid_h = NULL;

bool tb_srv_init(
        su_inifile_t* inifile,
        dbe_cryptoparams_t* cp)
{
        int i;
        bool succp;
        SscRetT sscret;
        char* args[7];
        int argsc;

        ss_pprintf_1(("tb_srv_init\n"));
        SS_PUSHNAME("tb_srv_init");

        if (!solid_started) {
            args[0]="";
            args[1] = "-Umysql";
            args[2] = "-Pmysql";
            args[3] = "-Cmysql";
            args[4] = "-n";
            args[5] = "solidDB for MySQL/Full functionality";
            args[6] = NULL;
            argsc = 6;
            solid_h = NULL;
            tb_init_inifile = inifile;
            ss_setservername("solidDB for MySQL/Full functionality");
            ss_pprintf_1(("tb_srv_init:SSCStartServer\n"));
            sscret = SSCStartServer(argsc, args, &solid_h, SSC_STATE_OPEN);
            ss_pprintf_1(("tb_srv_init:SSCStartServer:rc %d\n", sscret));
            solid_started = (sscret == SSC_SUCCESS);
            if (solid_started) {
                /* Create a global mutex for relation cursor list access.
                 * TODO: Semnum is not exactly correct but works ok here.
                 */
                ss_setservername("solidDB for MySQL/Full functionality");
                SsPrintf("%s\n", "solidDB for MySQL/Full functionality");
            }
            srv_shutdown = FALSE;
        }
        SS_POPNAME;
        return(solid_started);
}

void tb_srv_done(void)
{
        ss_pprintf_1(("tb_srv_done\n"));

        srv_shutdown = TRUE;

        if (solid_started) {
            ss_pprintf_1(("Stop solid server...\n"));
            if (SSCStopServer(solid_h, FALSE) != SSC_SUCCESS) {
                ss_pprintf_1(("Stop solid Server (force)...\n"));
                SSCStopServer(solid_h, TRUE);
            }
            solid_started = 0;
            solid_h = NULL;
        }
        ss_pprintf_2(("tb_srv_done:return\n"));
}

bool tb_srv_start(tb_database_t* tdb)
{
        ss_pprintf_1(("tb_srv_start\n"));
        return(TRUE);
}

void tb_srv_stop(void)
{
        ss_pprintf_1(("tb_srv_stop\n"));
}

#endif /* SS_MYSQL */

static void srv_lockwakeup(void* task)
{
        srv_lockwait_t* lw = task;

        CHK_LW(lw);
        ss_pprintf_1(("srv_lockwakeup\n"));

        SsMesSend(lw->lw_mes);
}

static bool srv_lockwait(void* task)
{
        srv_lockwait_t* lw = task;
        long timeout;

        CHK_LW(lw);
        ss_pprintf_1(("srv_lockwait:timeout=%ld\n", lw->lw_timeout));

        timeout = lw->lw_timeout;
        lw->lw_timeout = 0;
        rs_sysi_setlockwaitfun(lw->lw_cd, NULL);

        SsMesRequest(lw->lw_mes, timeout);

        ss_pprintf_1(("srv_lockwait:wake up\n"));

        return(TRUE);
}

static void srv_startlockwait(void* task)
{
        srv_lockwait_t* lw = task;

        CHK_LW(lw);
        ss_pprintf_1(("srv_startlockwait\n"));

        lw->lw_timeout = SSSEM_INDEFINITE_WAIT;

        rs_sysi_setlockwaitfun(lw->lw_cd, srv_lockwait);
}

static void srv_startlockwait_timeout(void* task, long timeout_ms)
{
        srv_lockwait_t* lw = task;

        CHK_LW(lw);
        ss_pprintf_1(("srv_startlockwait_timeout:timeout_ms=%ld\n", timeout_ms));

        lw->lw_timeout = timeout_ms;

        rs_sysi_setlockwaitfun(lw->lw_cd, srv_lockwait);
}

void tb_srv_initcd(rs_sysi_t* cd)
{
        srv_lockwait_t* lw;

        ss_pprintf_1(("tb_srv_initcd\n"));

        lw = SSMEM_NEW(srv_lockwait_t);

        ss_debug(lw->lw_chk = TBCHK_SRVLOCKWAIT);
        lw->lw_mes = SsMesCreateLocal();
        lw->lw_cd = cd;
        lw->lw_timeout = 0;
        CHK_LW(lw);

        rs_sysi_settask(cd, lw);
        rs_sysi_setlockwakeupfun(cd, srv_lockwakeup);
        rs_sysi_setstartlockwaitfun(cd, srv_startlockwait, srv_startlockwait_timeout);
}

void tb_srv_donecd(rs_sysi_t* cd)
{
        srv_lockwait_t* lw;

        ss_pprintf_1(("tb_srv_donecd\n"));

        lw = rs_sysi_task(cd);
        CHK_LW(lw);
        ss_dassert(lw->lw_timeout == 0);

        SsMesFree(lw->lw_mes);
        SsMemFree(lw);
}


/* ---------------------------------------------------------------------
 * performance output stuff
 */
typedef struct {
        SS_FILE*       pd_fp;
        uint           pd_interval;
        bool           pd_names;
        bool           pd_diff;
        int            pd_userid;
        bool           pd_raw;
        char*          pd_comment;
        long           pd_activectr;
        tb_database_t* pd_tdb;
} pmon_difftask_t;

static long pmon_difftask_activectr = 0;
static bool pmon_difftask_active = FALSE;

static void pmon_print(pmon_difftask_t* pd, bool raw, bool names, bool diff)
{
        char** filters = NULL;
        static ss_pmon_t last_pmon;
        static bool is_last_pmon;
        SsTimeT time_ms;
        static SsTimeT last_time_ms;
        bool print_timediff = TRUE;
        bool print_time = FALSE;
        SsTimeT timediff_ms;

        if (raw || names || diff) {
            /* Raw, name or diff output.
             */
            int i;
            int j;
            ss_pmon_t* pmon;
            dstr_t ds = NULL;
            bool firstp = TRUE;
            char buf[80];

            filters = (char**)SsMemAlloc(sizeof(filters[0]) * 2);
            filters[0] = "";
            filters[1] = NULL;

            pmon = SSMEM_NEW(ss_pmon_t);

            tb_database_pmonupdate_nomutex(pd->pd_tdb);
            SsPmonGetData(pmon);
            if (names || diff) {
                time_ms = SsTimeMs();
                timediff_ms = time_ms - last_time_ms;
                last_time_ms = time_ms;
            }

            dstr_set(&ds, "");

            if (diff && !is_last_pmon) {
                memcpy(&last_pmon, pmon, sizeof(ss_pmon_t));
                is_last_pmon = TRUE;
            }
            if (print_time) {
                if (names) {
                    SsSprintf(buf, "Time");
                } else {
                    SsSprintf(buf, "%04ld", SsTime(NULL)%10000);
                }
                dstr_app(&ds, buf);
                firstp = FALSE;
            }
            if (print_timediff) {
                if (names) {
                    SsSprintf(buf, "TimeMs");
                } else {
                    SsSprintf(buf, "%ld", timediff_ms);
                }
                dstr_app(&ds, buf);
                firstp = FALSE;
            }
            for (j = 0; filters[j] != NULL; j++) {
                for (i = 0; i < SS_PMON_MAXVALUES; i++) {
                    if (SsPmonAccept(filters[j], (ss_pmon_val_t)i)) {
                        if (names) {
                            SsSprintf(buf, "%s%s", firstp ? "" : ",", ss_pmon_names[i].pn_name);
                            memcpy(&last_pmon, pmon, sizeof(ss_pmon_t));
                            is_last_pmon = TRUE;
                        } else if (diff) {
                            if (ss_pmon_names[i].pn_type == SS_PMONTYPE_VALUE) {
                                SsSprintf(buf, "%s%lu",  firstp ? "" : ",", pmon->pm_values[i]);
                            } else {
                                long valdiff;
                                valdiff = pmon->pm_values[i] - last_pmon.pm_values[i];
                                if (raw) {
                                    SsSprintf(buf, "%s%lu",  firstp ? "" : ",", valdiff);
                                } else {
                                    double val;
                                    double secs;
                                    secs = (double)timediff_ms / 1000.0;
                                    if (secs < 0.0001) {
                                        val = 0.0;
                                    } else {
                                        val = valdiff / secs;
                                    }
                                    SsSprintf(buf, "%s%.2lf",  firstp ? "" : ",", val);
                                }
                            }
                        } else {
                            SsSprintf(buf, "%lu ", pmon->pm_values[i]);
                        }
                        dstr_app(&ds, buf);
                        firstp = FALSE;
                    }
                }
            }
            if (diff) {
                memcpy(&last_pmon, pmon, sizeof(ss_pmon_t));
            }

            SsFPrintf(pd->pd_fp, "%s\n", ds);

            dstr_free(&ds);
            SsMemFree(pmon);
            SsMemFree(filters);
        }
}

static void pmon_difftask(void* ctx, SsTimerRequestIdT req_id)
{
        pmon_difftask_t* pd = (pmon_difftask_t*)ctx;
        char* str = "pmon_difftask";
        SsTimerRequestIdT tid = 0;
        bool diff = TRUE;

        ss_dprintf_2(("pmon_difftask:tid %ld\n", (long)req_id));

        if (!SsTimerRequestIsValid(req_id)) {
            return;
        }
        if (srv_shutdown || pmon_difftask_activectr != pd->pd_activectr) {
            SsTimerCancelRequest(req_id);
            su_usrid_done(pd->pd_userid);
            if (pd->pd_fp != NULL) {
                SsFClose(pd->pd_fp);
            }
            if (pd->pd_comment != NULL) {
                SsMemFree(pd->pd_comment);
            }
            SsMemFree(pd);
            return;
        }

        pmon_print(pd, pd->pd_raw, pd->pd_names, pd->pd_diff);
        pd->pd_names = FALSE;
        pd->pd_diff = TRUE;
        tid = SsTimerAddRequest(pd->pd_interval, pmon_difftask, pd);
}


bool tb_srv_pmondiff_start(
        rs_sysi_t* cd,
        uint interval,
        char* filename,
        bool append,
        bool raw,
        char* comment,
        su_err_t** p_errh)
{
        bool succp = TRUE;
        pmon_difftask_t* pd;

        if (pmon_difftask_active) {
            su_err_init(p_errh, DBE_ERR_FAILED);
            return(FALSE);
        }

        pd = (pmon_difftask_t*)SsMemCalloc(sizeof(pmon_difftask_t), 1);
        pd->pd_activectr = pmon_difftask_activectr;
        pd->pd_names = TRUE;
        pd->pd_diff = FALSE;
        pd->pd_tdb = (tb_database_t*)rs_sysi_tabdb(cd);
        pd->pd_comment = NULL;
        pd->pd_interval = interval;
        pd->pd_raw = raw;

        pd->pd_userid = su_usrid_init();
        if (append) {
            pd->pd_fp = SsFOpenT(filename, "a+");
        } else {
            pd->pd_fp = SsFOpenT(filename, "w");
        }
        if (pd->pd_fp == NULL) {
            succp = FALSE;
            su_err_init(p_errh, DBE_ERR_FAILED);
        } else {
            pmon_difftask_active = TRUE;
            SsTimerAddRequest(pd->pd_interval, pmon_difftask, pd);
        }

        return(succp);
}

bool tb_srv_pmondiff_stop(
        rs_sysi_t* cd,
        su_err_t** p_errh)
{
        bool succp = TRUE;

        pmon_difftask_activectr++;
        pmon_difftask_active = FALSE;

        return(succp);
}

void tb_srv_pmondiff_clear(
        rs_sysi_t* cd)
{
        SsPmonClear();
        /* sse_admin_pmonclear(); */
        ss_beta(dbe_logi_clearlogrecinfo();)
}

